import numpy as np
import pandas as pd
import polars as pl
import pytest
from great_tables import GT, google_font, loc, style

from gt_extras.formatting import (
    GTCombinedLayout,
    fmt_pct_extra,
    gt_duplicate_column,
    gt_two_column_layout,
)
from gt_extras.tests.conftest import assert_rendered_body


def test_fmt_pct_extra_snap(snapshot, mini_gt):
    res = fmt_pct_extra(mini_gt, columns="num", scale=1, decimals=0)
    assert_rendered_body(snapshot, gt=res)


def test_fmt_pct_extra_basic(mini_gt):
    html = fmt_pct_extra(mini_gt, columns="num", scale=1).as_raw_html()

    assert "<span style='color:grey;'><1%</span>" in html
    assert "2.2%" in html
    assert "33.3%" in html


def test_fmt_pct_extra_threshold_low(mini_gt):
    html = fmt_pct_extra(mini_gt, columns="num", scale=100, threshold=10).as_raw_html()

    assert "11.1%" in html
    assert "222.2%" in html
    assert "3333.0%" in html


def test_fmt_pct_extra_threshold_high(mini_gt):
    html = fmt_pct_extra(
        mini_gt, columns="num", scale=100, threshold=4000
    ).as_raw_html()

    assert html.count("<span style='color:grey;'><4000%</span>") == 3


def test_fmt_pct_extra_custom_color(mini_gt):
    html = fmt_pct_extra(
        mini_gt, columns="num", color="red", threshold=50
    ).as_raw_html()

    assert html.count("<span style='color:red;'><50%</span>") == 1


def test_fmt_pct_extra_decimals(mini_gt):
    html_0 = fmt_pct_extra(mini_gt, columns="num", decimals=0).as_raw_html()
    assert "11%" in html_0  # 0.1111 * 100 = 11.11% rounded
    assert "222%" in html_0


def test_fmt_pct_extra_negative_values():
    df = pd.DataFrame({"num": [-0.005, -0.25, 0.15]})
    gt_test = GT(df)

    html = fmt_pct_extra(gt=gt_test, columns="num", threshold=1.0).as_raw_html()

    assert "<span style='color:grey;'><1%</span>" in html
    assert "-25.0%" in html
    assert "15.0%" in html


def test_fmt_pct_extra_zero_values():
    df = pd.DataFrame({"num": [0.0, 0.005, 0.02]})
    gt_test = GT(df)
    html = fmt_pct_extra(gt=gt_test, columns="num").as_raw_html()

    assert html.count("<span style='color:grey;'><1%</span>") == 2
    assert "2.0%" in html


def test_fmt_pct_extra_edge_case_threshold():
    df = pd.DataFrame({"num": [0.01, 0.0099, 0.0101]})
    gt_test = GT(df)

    html = fmt_pct_extra(
        gt=gt_test, columns="num", scale=100, threshold=1.0, decimals=2
    ).as_raw_html()

    assert "1.00%" in html
    assert "<span style='color:grey;'><1%</span>" in html
    assert "1.01%" in html


def test_fmt_pct_extra_with_none_values():
    df = pd.DataFrame({"num": [0.005, None, 0.25, np.nan]})
    gt_test = GT(df)

    result = fmt_pct_extra(gt=gt_test, columns="num")
    html = result.as_raw_html()

    assert isinstance(result, GT)
    assert "25%" in html


def test_gt_duplicate_column_snap(snapshot, mini_gt):
    res = gt_duplicate_column(mini_gt, column="num")
    assert_rendered_body(snapshot, gt=res)


def test_gt_duplicate_column_basic(mini_gt):
    res = gt_duplicate_column(mini_gt, column="num", append_text="_copy")
    html = res.as_raw_html()

    assert "num_copy" in res._tbl_data.columns
    assert "num_copy" in html
    assert all(res._tbl_data["num"] == res._tbl_data["num_copy"])


def test_gt_duplicate_column_custom_name(mini_gt):
    res = gt_duplicate_column(mini_gt, column="num", dupe_name="duplicated_num")
    html = res.as_raw_html()

    assert "duplicated_num" in res._tbl_data.columns
    assert "duplicated_num" in html
    assert all(res._tbl_data["num"] == res._tbl_data["duplicated_num"])


def test_gt_duplicate_column_position(mini_gt):
    res = gt_duplicate_column(mini_gt, column="num", after="char")
    html = res.as_raw_html()

    assert "num_dupe" in res._tbl_data.columns
    assert "num_dupe" in html

    columns = list(res._boxhead)
    assert columns[2].column_label == "num_dupe"


def test_gt_duplicate_column_polars():
    df = pl.DataFrame({"num": [1, 2, 3], "char": ["a", "b", "c"]})
    gt_test = GT(df)

    res = gt_duplicate_column(gt_test, column="num", append_text="_copy")
    html = res.as_raw_html()

    assert "num_copy" in res._tbl_data.columns
    assert "num_copy" in html

    original_values = res._tbl_data.get_column("num").to_list()  # type: ignore
    duplicated_values = res._tbl_data.get_column("num_copy").to_list()  # type: ignore
    assert original_values == duplicated_values


def test_gt_duplicate_column_invalid_name(mini_gt):
    with pytest.raises(
        ValueError, match="cannot be the same as the original column name"
    ):
        gt_duplicate_column(mini_gt, column="num", dupe_name="num")


@pytest.fixture
def two_dfs():
    df1 = pd.DataFrame({"A": [1, 2], "B": [3, 4]})
    df2 = pd.DataFrame({"A": [5, 6], "B": [7, 8]})
    return df1, df2


def test_gt_two_column_layout_snapshot(snapshot, two_dfs):
    df1, df2 = two_dfs
    gt1 = GT(df1, id="id1").tab_header(title="Header 1", subtitle="Subtitle 1")
    gt2 = GT(df2, id="id2").tab_header(title="Header 2", subtitle="Subtitle 2")

    result = gt_two_column_layout(gt1, gt2, table_header_from=1)
    assert snapshot == str(result)


def test_gt_two_column_layout_basic(two_dfs):
    df1, df2 = two_dfs
    gt1 = GT(df1).tab_header(title="Left Table", subtitle="Left Subtitle")
    gt2 = GT(df2).tab_header(title="Right Table", subtitle="Right Subtitle")

    result = gt_two_column_layout(gt1, gt2)
    html = str(result)

    assert "Left Table" in html
    assert "Right Table" in html
    assert '<div id="mycombinedtable"' in html
    assert "<table" in html


def test_gt_two_column_layout_with_header_from_1(two_dfs):
    df1, df2 = two_dfs
    gt1 = GT(df1).tab_header(title="Header 1", subtitle="Subtitle 1")
    gt2 = GT(df2).tab_header(title="Header 2", subtitle="Subtitle 2")

    result = gt_two_column_layout(gt1, gt2, table_header_from=1)
    html = str(result)

    assert "Header 1" in html
    assert "Subtitle 1" in html
    assert "Header 2" not in html
    assert "Subtitle 2" not in html


def test_gt_two_column_layout_with_header_from_2(two_dfs):
    df1, df2 = two_dfs
    gt1 = GT(df1).tab_header(title="Header 1", subtitle="Subtitle 1")
    gt2 = GT(df2).tab_header(title="Header 2", subtitle="Subtitle 2")

    result = gt_two_column_layout(gt1, gt2, table_header_from=2)
    html = str(result)

    assert "Header 2" in html
    assert "Subtitle 2" in html
    assert "Header 1" not in html
    assert "Subtitle 1" not in html


def test_gt_two_column_layout_no_headers(two_dfs):
    df1, df2 = two_dfs
    gt1 = GT(df1)
    gt2 = GT(df2)

    result = gt_two_column_layout(gt1, gt2, table_header_from=1)
    html = str(result)

    assert "Header" not in html
    assert "Subtitle" not in html


def test_gt_two_column_layout_with_styles(two_dfs):
    df1, df2 = two_dfs
    gt1 = GT(df1)
    gt2 = GT(df2)

    gt1 = (
        GT(df1)
        .tab_header(title="Table 1", subtitle="1st Table")
        .tab_style(
            style=style.text(font=google_font("Chivo"), weight="bold"),
            locations=loc.title(),
        )
        .tab_style(
            style=style.text(font=google_font("Chivo")),
            locations=loc.subtitle(),
        )
    )

    gt2 = GT(df2).tab_header(title="Table 2", subtitle="2nd Table")

    # Create combined layout using table 1's header
    result = gt_two_column_layout(gt1, gt2, table_header_from=1)

    # Get the HTML output
    html = str(result)

    assert 'style="width:100%; font-family: Chivo;"' in html
    assert 'style="width:100%; font-family: Chivo;font-weight: bold;"' in html


def test_gt_combined_layout_repr_html():
    html = "<div>hello world</div>"
    layout = GTCombinedLayout(html)
    assert layout._repr_html_() == html


def test_gt_two_column_layout_save_target(two_dfs):
    df1, df2 = two_dfs
    gt1 = GT(df1)
    gt2 = GT(df2)
    with pytest.raises(
        NotImplementedError,
        match="At the moment, only notebook and browser display options are available.",
    ):
        gt_two_column_layout(gt1, gt2, target="save")  # type: ignore


def test_gt_two_column_layout_invalid_target(two_dfs):
    df1, df2 = two_dfs
    gt1 = GT(df1)
    gt2 = GT(df2)
    with pytest.raises(Exception, match="Unknown target display"):
        gt_two_column_layout(gt1, gt2, target="invalid")  # type: ignore


def test_gt_combined_layout_save_incomplete(two_dfs):
    df1, df2 = two_dfs
    gt1 = GT(df1)
    gt2 = GT(df2)
    gt_combined = gt_two_column_layout(gt1, gt2)

    with pytest.raises(NotImplementedError):
        gt_combined.save()


@pytest.mark.xfail(reason="Notebook target test not implemented yet")
def test_gt_two_column_layout_notebook_target():
    assert False


@pytest.mark.xfail(reason="Browser target test not implemented yet")
def test_gt_two_column_layout_browser_target():
    assert False



import numpy as np
import pandas as pd
import pytest
from great_tables import GT, loc, style

from gt_extras import (
    gt_plt_bar,
    gt_plt_bar_pct,
    gt_plt_bar_stack,
    gt_plt_bullet,
    gt_plt_conf_int,
    gt_plt_donut,
    gt_plt_dot,
    gt_plt_dumbbell,
    gt_plt_winloss,
)
from gt_extras.tests.conftest import assert_rendered_body


def test_gt_plt_bar_snap(snapshot, mini_gt):
    res = gt_plt_bar(gt=mini_gt, columns="num")

    assert_rendered_body(snapshot, gt=res)


def test_gt_plt_bar(mini_gt):
    result = gt_plt_bar(gt=mini_gt, columns=["num"])
    html = result.as_raw_html()
    assert html.count("<svg") == 3


def test_gt_plt_bar_bar_height_too_high(mini_gt):
    with pytest.warns(
        UserWarning,
        match="Bar_height must be less than or equal to the plot height. Adjusting bar_height to 567.",
    ):
        result = gt_plt_bar(gt=mini_gt, columns=["num"], bar_height=1234, height=567)
        html = result.as_raw_html()

    assert html.count('height="567"') == 3
    assert 'height="1234"' not in html


def test_gt_plt_bar_bar_height_too_low(mini_gt):
    with pytest.warns(
        UserWarning,
        match="Bar_height cannot be negative. Adjusting bar_height to 0.",
    ):
        result = gt_plt_bar(gt=mini_gt, columns=["num"], bar_height=-345, height=1234)
        html = result.as_raw_html()

    assert html.count('height="1234"') == 3
    assert 'height="-345"' not in html


def test_gt_plt_bar_show_labels_true(mini_gt):
    result = gt_plt_bar(gt=mini_gt, columns=["num"], show_labels=True)
    html = result.as_raw_html()
    assert ">33.33</text>" in html


def test_gt_plt_bar_keep_columns(mini_gt):
    gt = mini_gt.tab_style(
        style=style.fill("lightblue"),
        locations=loc.body(),
    )
    result = gt_plt_bar(gt=gt, columns=["num"], keep_columns=True)
    html = result.as_raw_html()

    assert ">num plot</th>" in html
    assert ">num</th>" in html
    assert ">2.222</td>" in html
    assert html.count("<svg") == 3


def test_gt_plt_bar_show_labels_false(mini_gt):
    result = gt_plt_bar(gt=mini_gt, columns=["num"], show_labels=False)
    html = result.as_raw_html()
    assert "</text>" not in html


def test_gt_plt_bar_no_stroke_color(mini_gt):
    result = gt_plt_bar(gt=mini_gt, columns=["num"], stroke_color=None)
    html = result.as_raw_html()
    assert html.count('line stroke="transparent"') == 3


def test_gt_plt_bar_type_error(mini_gt):
    with pytest.raises(TypeError, match="Invalid column type provided"):
        gt_plt_bar(gt=mini_gt, columns=["char"])


def test_gt_plt_dot_snap(snapshot, mini_gt):
    res = gt_plt_dot(gt=mini_gt, category_col="fctr", data_col="currency")

    assert_rendered_body(snapshot, gt=res)


def test_gt_plt_dot_basic(mini_gt):
    result = gt_plt_dot(gt=mini_gt, category_col="char", data_col="num")
    html = result.as_raw_html()

    assert html.count("<svg") == 3
    assert html.count('<circle cx="5.818181818181818" cy="15.0"') == 3
    assert html.count('<text dominant-baseline="central"') == 3
    assert html.count('text-anchor="start" font-size="16"') == 3
    assert html.count('<rect x="0"') == 3


def test_gt_plt_dot_with_palette(mini_gt):
    result = gt_plt_dot(
        gt=mini_gt,
        category_col="char",
        data_col="num",
        palette=["#FF0000", "#00FF00", "#0000FF"],
    )
    html = result.as_raw_html()

    assert "#ff0000" in html
    assert "#00ff00" in html
    assert "#0000ff" in html


def test_gt_plt_dot_with_domain_expanded(mini_gt):
    result = gt_plt_dot(
        gt=mini_gt, category_col="char", data_col="num", domain=[0, 100]
    )
    html = result.as_raw_html()

    assert '<rect x="0" y="24.6" width="39.995999999999995"' in html
    assert '<rect x="0" y="24.6" width="2.6664"' in html
    assert '<rect x="0" y="24.6" width="0.13332"' in html


def test_gt_plt_dot_with_domain_restricted(mini_gt):
    with pytest.warns(
        UserWarning,
        match="Value 33.33 in column 'num' is greater than the domain maximum 10. Setting to 10.",
    ):
        result = gt_plt_dot(
            gt=mini_gt, category_col="char", data_col="num", domain=[0, 10]
        )
        html = result.as_raw_html()

    assert '<rect x="0" y="24.6" width="1.3332"' in html
    assert '<rect x="0" y="24.6" width="26.664"' in html
    assert '<rect x="0" y="24.6" width="120"' in html


def test_gt_plt_dot_invalid_data_col(mini_gt):
    with pytest.raises(KeyError, match="Column 'invalid_col' not found"):
        gt_plt_dot(gt=mini_gt, category_col="char", data_col="invalid_col")


def test_gt_plt_dot_invalid_category_col(mini_gt):
    with pytest.raises(KeyError, match="Column 'invalid_col' not found"):
        gt_plt_dot(gt=mini_gt, category_col="invalid_col", data_col="num")


def test_gt_plt_dot_multiple_data_cols(mini_gt):
    with pytest.raises(
        ValueError, match="Expected a single column, but got multiple columns"
    ):
        gt_plt_dot(gt=mini_gt, category_col="char", data_col=["num", "char"])


def test_gt_plt_dot_multiple_category_cols(mini_gt):
    with pytest.raises(
        ValueError, match="Expected a single column, but got multiple columns"
    ):
        gt_plt_dot(gt=mini_gt, category_col=["char", "num"], data_col="num")


def test_gt_plt_dot_non_numeric_data_col(mini_gt):
    with pytest.raises(TypeError, match="Invalid column type provided"):
        gt_plt_dot(gt=mini_gt, category_col="char", data_col="char")


def test_gt_plt_dot_with_na_values():
    df = pd.DataFrame(
        {
            "category": ["A", "B", "C", "D"],
            "values": [10, np.nan, 20, None],
        }
    )
    gt = GT(df)

    result = gt_plt_dot(gt=gt, category_col="category", data_col="values")
    html = result.as_raw_html()

    assert isinstance(result, GT)
    assert '<rect x="0" y="24.6" width="60.0"' in html
    assert '<rect x="0" y="24.6" width="120.0"' in html
    assert html.count('<rect x="0" y="24.6" width="0"') == 2


def test_gt_plt_dot_with_na_in_category():
    df = pd.DataFrame(
        {
            "category": [np.nan, "B", None, None],
            "values": [5, 10, 10, 5],
        }
    )
    gt = GT(df)

    with pytest.warns(UserWarning, match="A color value is None and has been coerced"):
        result = gt_plt_dot(gt=gt, category_col="category", data_col="values")

    html = result.as_raw_html()

    assert isinstance(result, GT)
    assert html.count('<rect x="0" y="24.6" width="120.0"') == 1
    assert '<rect x="0" y="24.6" width="60.0"' not in html


def test_gt_plt_dot_palette_string_valid(mini_gt):
    result = gt_plt_dot(
        gt=mini_gt, category_col="char", data_col="num", palette="viridis"
    )
    html = result.as_raw_html()

    assert 'fill="#440154"' in html
    assert 'fill="#22908c"' in html
    assert 'fill="#fde725"' in html


def test_gt_plt_conf_int_snap(snapshot):
    df = pd.DataFrame(
        {
            "group": ["A", "B", "C"],
            "mean": [5.2, 7.8, 3.4],
            "ci_lower": [4.1, 6.9, 2.8],
            "ci_upper": [6.3, 8.7, 4.0],
        }
    )
    gt_test = GT(df)
    res = gt_plt_conf_int(
        gt=gt_test, column="mean", ci_columns=["ci_lower", "ci_upper"]
    )

    assert_rendered_body(snapshot, gt=res)


def test_gt_plt_conf_int_basic():
    df = pd.DataFrame(
        {
            "group": ["A", "B", "C"],
            "mean": [1, 2, 3],
            "ci_lower": [0, 2, 2],
            "ci_upper": [4, 6, 5],
        }
    )
    gt_test = GT(df)
    result = gt_plt_conf_int(
        gt=gt_test, column="mean", ci_columns=["ci_lower", "ci_upper"]
    )
    html = result.as_raw_html()

    assert html.count("<svg") == 3
    assert html.count('height="3.0" rx="2" fill="royalblue"') == 3
    assert html.count('<circle stroke="red" stroke-width="1.5"') == 3
    assert html.count('cy="20.0" r="3.0" fill="red"') == 3

    assert html.count('<text dominant-baseline="central"') == 6
    assert html.count('text-anchor="start" font-size="10" fill="black"') == 3
    assert html.count('text-anchor="end" font-size="10" fill="black"') == 3

    assert 'cx="22.222222222222225"' in html
    assert 'cx="36.111111111111114"' in html
    assert 'cx="50.000000000000014"' in html

    assert '<rect x="8.333333333333336"' in html
    assert '<rect x="36.111111111111114"' in html
    assert 'width="55.55555555555556"' in html
    assert 'width="41.66666666666667"' in html

    assert 'x="8.333333333333336" y="12.5">0</text>' in html
    assert 'x="63.88888888888889" y="12.5">4</text>' in html
    assert 'x="36.111111111111114" y="12.5">2</text>' in html
    assert 'x="91.66666666666667" y="12.5">6</text>' in html


def test_gt_plt_conf_int_computed_ci():
    df = pd.DataFrame(
        {
            "group": ["A", "B"],
            "data": [[1, 2, 2, 5, 6] * 5, [1, 5, 5, 9] * 10],
        }
    )
    gt_test = GT(df)
    result = gt_plt_conf_int(gt=gt_test, column="data")
    html = result.as_raw_html()

    assert ">2.4</text>" in html
    assert ">4</text>" in html
    assert ">4.1</text>" in html
    assert ">5.9</text>" in html


def test_gt_plt_conf_int_custom_colors():
    df = pd.DataFrame(
        {
            "group": ["A", "B"],
            "mean": [5.2, 7.8],
            "ci_lower": [4.1, 6.9],
            "ci_upper": [6.3, 8.7],
        }
    )
    gt_test = GT(df)
    result = gt_plt_conf_int(
        gt=gt_test,
        column="mean",
        ci_columns=["ci_lower", "ci_upper"],
        line_color="blue",
        dot_color="green",
        text_color="red",
    )
    html = result.as_raw_html()

    assert html.count('fill="blue"') == 2
    assert html.count('fill="green"') == 2
    assert html.count('fill="red"') == 4


def test_gt_plt_conf_int_invalid_column():
    df = pd.DataFrame(
        {
            "group": ["A", "B"],
            "mean": [5.2, 7.8],
            "ci_lower": [4.1, 6.9],
            "ci_upper": [6.3, 8.7],
        }
    )
    gt_test = GT(df)

    with pytest.raises(
        ValueError, match="Expected a single column, but got multiple columns"
    ):
        gt_plt_conf_int(gt=gt_test, column=["mean", "group"])


def test_gt_plt_conf_int_invalid_ci_columns():
    df = pd.DataFrame(
        {
            "group": ["A", "B"],
            "mean": [5.2, 7.8],
            "ci_lower": [4.1, 6.9],
            "ci_upper": [6.3, 8.7],
        }
    )
    gt_test = GT(df)

    with pytest.raises(ValueError, match="Expected 2 ci_columns"):
        gt_plt_conf_int(gt=gt_test, column="mean", ci_columns=["ci_lower"])


def test_gt_plt_conf_int_with_none_values():
    df = pd.DataFrame(
        {
            "group": ["A", "B", "C"],
            "mean": [5.2, None, 3.4],
            "ci_lower": [4.1, None, 2.8],
            "ci_upper": [6.3, np.nan, 4.0],
        }
    )
    gt_test = GT(df)
    result = gt_plt_conf_int(
        gt=gt_test, column="mean", ci_columns=["ci_lower", "ci_upper"]
    )

    assert isinstance(result, GT)

    html = result.as_raw_html()

    assert '<div style="width:100px; height:30px;"></div>' in html


def test_gt_plt_conf_int_computed_invalid_data():
    df = pd.DataFrame(
        {
            "group": ["A", "B"],
            "data": [5.2, 7.8],  # Not lists
        }
    )
    gt_test = GT(df)

    with pytest.raises(
        ValueError, match="Expected entries in data to be lists or None"
    ):
        gt_plt_conf_int(gt=gt_test, column="data")


def test_gt_plt_conf_int_empty_data():
    df = pd.DataFrame(
        {
            "group": ["A", "B"],
            "data": [[], [1, 2, 3, 4, 5, 6]],
        }
    )
    gt_test = GT(df)
    result = gt_plt_conf_int(gt=gt_test, column="data")
    html = result.as_raw_html()

    assert html.count("<circle") == 1
    assert html.count("<rect") == 1


def test_gt_plt_conf_int_precomputed_invalid_data():
    df = pd.DataFrame(
        {
            "group": ["A", "B"],
            "mean": [["not", "numeric"], [7.8]],  # Not numeric
            "ci_lower": [4.1, 6.9],
            "ci_upper": [6.3, 8.7],
        }
    )
    gt_test = GT(df)

    with pytest.raises(
        ValueError, match="Expected all entries in mean to be numeric or None"
    ):
        gt_plt_conf_int(gt=gt_test, column="mean", ci_columns=["ci_lower", "ci_upper"])


def test_gt_plt_dumbbell_snap(snapshot):
    df = pd.DataFrame({"value_1": [10, 15, 25], "value_2": [15, 20, 30]})
    gt_test = GT(df)
    res = gt_plt_dumbbell(gt=gt_test, col1="value_1", col2="value_2")

    assert_rendered_body(snapshot, gt=res)


def test_gt_plt_dumbbell_basic():
    df = pd.DataFrame({"value_1": [10, 15, 25], "value_2": [15, 20, 30]})
    gt_test = GT(df)
    result = gt_plt_dumbbell(gt=gt_test, col1="value_1", col2="value_2")
    html = result.as_raw_html()

    assert html.count('<circle stroke="white" stroke-width="1.5"') == 6
    assert html.count('r="3.75" fill="purple"') == 3
    assert html.count('r="3.75" fill="green"') == 3

    assert html.count("<rect") == 3
    assert html.count('height="3.0" rx="2" fill="grey"') == 3

    assert 'x="8.333333333333332" y="18.5" width="20.833333333333336"' in html
    assert 'x="29.166666666666668" y="18.5" width="20.833333333333332"' in html
    assert 'x="70.83333333333334" y="18.5" width="20.833333333333314"' in html

    assert html.count('<text dominant-baseline="lower" text-anchor="middle"') == 6
    assert html.count('font-size="10" font-weight="bold" fill="green"') == 3
    assert html.count('font-size="10" font-weight="bold" fill="purple"') == 3

    assert ">10</text>" in html
    assert ">15</text>" in html
    assert ">20</text>" in html
    assert ">25</text>" in html
    assert ">30</text>" in html


def test_gt_plt_dumbbell_custom_colors():
    df = pd.DataFrame({"group": ["A", "B"], "value_1": [10, 20], "value_2": [15, 25]})
    gt_test = GT(df)

    result = gt_plt_dumbbell(
        gt=gt_test,
        col1="value_1",
        col2="value_2",
        col1_color="blue",
        col2_color="red",
        bar_color="green",
    )
    html = result.as_raw_html()

    assert html.count('fill="blue"') == 4
    assert html.count('fill="red"') == 4
    assert html.count('fill="green"') == 2


def test_gt_plt_dumbbell_custom_dimensions():
    df = pd.DataFrame({"group": ["A", "B"], "value_1": [10, 20], "value_2": [15, 25]})
    gt_test = GT(df)

    result = gt_plt_dumbbell(
        gt=gt_test, col1="value_1", col2="value_2", width=200, height=50
    )
    html = result.as_raw_html()

    assert '<svg xmlns="http://www.w3.org/2000/svg" width="200" height="50">' in html


def test_gt_plt_dumbbell_font_size():
    df = pd.DataFrame({"group": ["A", "B"], "value_1": [10, 20], "value_2": [15, 25]})
    gt_test = GT(df)

    result = gt_plt_dumbbell(gt=gt_test, col1="value_1", col2="value_2", font_size=14)
    html = result.as_raw_html()

    assert html.count('font-size="14"') == 4


def test_gt_plt_dumbbell_decimals():
    df = pd.DataFrame(
        {"group": ["A", "B"], "value_1": [10.123, 20.456], "value_2": [15.789, 25.012]}
    )
    gt_test = GT(df)

    result = gt_plt_dumbbell(gt=gt_test, col1="value_1", col2="value_2", num_decimals=2)
    html = result.as_raw_html()

    assert "10.12" in html
    assert "15.79" in html


def test_gt_plt_dumbbell_with_label():
    df = pd.DataFrame({"group": ["A", "B"], "value_1": [10, 20], "value_2": [15, 25]})
    gt_test = GT(df)

    result = gt_plt_dumbbell(
        gt=gt_test, col1="value_1", col2="value_2", label="Custom Label"
    )

    html = result.as_raw_html()
    assert "Custom Label" in html


def test_gt_plt_dumbbell_hides_col2():
    df = pd.DataFrame({"group": ["A", "B"], "value_1": [10, 20], "value_2": [15, 25]})
    gt_test = GT(df)
    result = gt_plt_dumbbell(gt=gt_test, col1="value_1", col2="value_2")
    html = result.as_raw_html()

    assert "value_2" not in html
    assert "value_1" in html
    assert "group" in html


def test_gt_plt_dumbbell_with_none_values():
    df = pd.DataFrame({"value_1": [10, None, 30], "value_2": [15, 25, None]})
    gt_test = GT(df)

    result = gt_plt_dumbbell(gt=gt_test, col1="value_1", col2="value_2")
    html = result.as_raw_html()

    assert html.count('<div style="width:100px; height:30px;"></div>') == 2


def test_gt_plt_dumbbell_with_na_values():
    df = pd.DataFrame({"value_1": [10, np.nan], "value_2": [np.nan, 25]})
    gt_test = GT(df)
    result = gt_plt_dumbbell(gt=gt_test, col1="value_1", col2="value_2")
    html = result.as_raw_html()

    assert html.count('<div style="width:100px; height:30px;"></div>') == 2


def test_gt_plt_dumbbell_invalid_col1():
    df = pd.DataFrame({"group": ["A", "B"], "value_1": [10, 20], "value_2": [15, 25]})
    gt_test = GT(df)

    with pytest.raises(KeyError):
        gt_plt_dumbbell(gt=gt_test, col1="invalid_col", col2="value_2")


def test_gt_plt_dumbbell_invalid_col2():
    df = pd.DataFrame({"value_1": [10, 20], "value_2": [15, 25]})
    gt_test = GT(df)

    with pytest.raises(KeyError):
        gt_plt_dumbbell(gt=gt_test, col1="value_1", col2="invalid_col")


def test_gt_plt_dumbbell_non_numeric_col1():
    df = pd.DataFrame({"value_1": ["text", "more_text"], "value_2": [15, 25]})
    gt_test = GT(df)

    with pytest.raises(ValueError, match="Expected all entries to be numeric or None."):
        gt_plt_dumbbell(gt=gt_test, col1="value_1", col2="value_2")


def test_gt_plt_dumbbell_non_numeric_col2():
    df = pd.DataFrame({"value_1": [10, 20], "value_2": ["123", 30]})
    gt_test = GT(df)

    with pytest.raises(ValueError, match="Expected all entries to be numeric or None."):
        gt_plt_dumbbell(gt=gt_test, col1="value_1", col2="value_2")


def test_gt_plt_dumbbell_same_values():
    df = pd.DataFrame({"value_1": [20, 20, 30], "value_2": [20, 30, 30]})
    gt_test = GT(df)

    result = gt_plt_dumbbell(gt=gt_test, col1="value_1", col2="value_2")
    html = result.as_raw_html()

    assert html.count('<text dominant-baseline="lower"') == 6
    assert html.count(">20</text>") == 3
    assert html.count(">30</text>") == 3
    assert html.count('width="0.0"') == 2


def test_gt_plt_dumbbell_reversed_values():
    df = pd.DataFrame({"value_1": [200, 300, 0], "value_2": [15, 20, 400]})
    gt_test = GT(df)

    result = gt_plt_dumbbell(gt=gt_test, col1="value_1", col2="value_2")
    html = result.as_raw_html()

    assert 'fill="purple" x="50.0" y="14.45">200</text>' in html
    assert 'fill="purple" x="70.83333333333334" y="14.45">300</text>' in html
    assert 'fill="purple" x="8.333333333333332" y="14.45">0</text>' in html

    assert 'fill="green" x="11.458333333333332" y="14.45">15</text>' in html
    assert 'fill="green" x="12.5" y="14.45">20</text>' in html
    assert 'fill="green" x="91.66666666666666" y="14.45">400</text>' in html


def test_gt_plt_winloss_snap(snapshot):
    df = pd.DataFrame(
        {
            "team": ["A", "B"],
            "games": [
                [1, 0.5, 0],
                [0, 0, 1],
            ],
        }
    )
    gt_test = GT(df)
    res = gt_plt_winloss(gt=gt_test, column="games")

    assert_rendered_body(snapshot, gt=res)


def test_gt_plt_winloss_basic():
    df = pd.DataFrame({"team": ["A", "B"], "games": [[1, 0, 0.5], [0, 1, 1]]})
    gt_test = GT(df)

    result = gt_plt_winloss(gt=gt_test, column="games")
    html = result.as_raw_html()

    assert html.count('y="6.0" width="24.666666666666668" ') == 3
    assert html.count('y="12.0" width="24.666666666666668" ') == 3

    assert html.count('height="12.0" rx="2"') == 5
    assert html.count('height="6.0" rx="2"') == 1

    assert html.count('<rect x="0.0"') == 2
    assert html.count('<rect x="26.666666666666668"') == 2
    assert html.count('<rect x="53.333333333333336"') == 2

    assert html.count('fill="grey"') == 1
    assert html.count('fill="red"') == 2
    assert html.count('fill="blue"') == 3


def test_gt_plt_winloss_custom_colors():
    df = pd.DataFrame({"team": ["A"], "games": [[1, 1, 1, 0, 0, 0.5]]})
    gt_test = GT(df)

    result = gt_plt_winloss(
        gt=gt_test,
        column="games",
        win_color="green",
        loss_color="black",
        tie_color="#FFA500",
    )
    html = result.as_raw_html()

    assert html.count('fill="green"') == 3
    assert html.count('fill="black"') == 2
    assert html.count('fill="#FFA500"') == 1


def test_gt_plt_winloss_custom_dimensions():
    df = pd.DataFrame({"team": ["A"], "games": [[1, 0]]})
    gt_test = GT(df)

    result = gt_plt_winloss(gt=gt_test, column="games", width=200, height=50)
    html = result.as_raw_html()

    assert '<svg xmlns="http://www.w3.org/2000/svg" width="200" height="50">' in html
    assert '<rect x="0.0" y="10.0" width="98.0" height="20.0" rx="2"' in html
    assert '<rect x="100.0" y="20.0" width="98.0" height="20.0" rx="2"' in html


def test_gt_plt_winloss_shape_square():
    df = pd.DataFrame({"team": ["A"], "games": [[1, 0]]})
    gt_test = GT(df)

    result = gt_plt_winloss(gt=gt_test, column="games", shape="square")
    html = result.as_raw_html()

    assert html.count('rx="0.5"') == 2


def test_gt_plt_winloss_shape_pill():
    df = pd.DataFrame({"team": ["A"], "games": [[1, 0]]})
    gt_test = GT(df)

    result = gt_plt_winloss(gt=gt_test, column="games", shape="pill")
    html = result.as_raw_html()

    assert html.count('rx="2"') == 2


def test_gt_plt_winloss_spacing():
    df = pd.DataFrame({"team": ["A"], "games": [[1, 0, 1]]})
    gt_test = GT(df)

    result = gt_plt_winloss(gt=gt_test, column="games", width=90, spacing=6)
    html = result.as_raw_html()

    assert '<rect x="0.0" y="6.0" width="24.0"' in html
    assert '<rect x="30.0" y="12.0" width="24.0"' in html
    assert '<rect x="60.0" y="6.0" width="24.0"' in html


def test_gt_plt_winloss_with_empty_list():
    df = pd.DataFrame({"team": ["A", "B"], "games": [[], [1, 0]]})
    gt_test = GT(df)

    result = gt_plt_winloss(gt=gt_test, column="games")
    html = result.as_raw_html()

    assert '<div style="width:80px; height:30px;"></div>' in html


def test_gt_plt_winloss_with_none_values():
    df = pd.DataFrame({"team": ["A", "B"], "games": [[np.nan, 1, None, 0], [0.5, 1]]})
    gt_test = GT(df)

    result = gt_plt_winloss(gt=gt_test, column="games")
    html = result.as_raw_html()

    assert html.count('fill="blue"') == 2
    assert html.count('fill="red"') == 1
    assert html.count('fill="grey"') == 1
    assert html.count('x="20.0"') == 2
    assert html.count('x="60.0"') == 1


def test_gt_plt_winloss_with_invalid_values():
    df = pd.DataFrame({"team": ["A"], "games": [[1, 0.2, 0.5, 2, 0, "invalid"]]})
    gt_test = GT(df)

    with pytest.warns(
        UserWarning, match="Invalid value '.*' encountered in win/loss data. Skipping."
    ):
        result = gt_plt_winloss(gt=gt_test, column="games")
        html = result.as_raw_html()

    assert html.count('fill="blue"') == 1
    assert html.count('fill="grey"') == 1
    assert html.count('fill="red"') == 1


def test_gt_plt_winloss_different_length_lists():
    df = pd.DataFrame({"team": ["A", "B"], "games": [[1, 0], [1, 0, 0.5, 1, 0]]})
    gt_test = GT(df)

    result = gt_plt_winloss(gt=gt_test, column="games")
    html = result.as_raw_html()

    assert html.count('x="16.0"') == 2
    assert html.count('x="32.0"') == 1

    assert html.count('fill="blue"') == 3
    assert html.count('fill="red"') == 3
    assert html.count('fill="grey"') == 1


def test_gt_plt_winloss_invalid_column():
    df = pd.DataFrame({"team": ["A"], "games": [[1, 0]]})
    gt_test = GT(df)

    with pytest.raises(KeyError):
        gt_plt_winloss(gt=gt_test, column="invalid_column")


def test_gt_plt_winloss_spacing_warning():
    df = pd.DataFrame({"team": ["A"], "games": [[1, 0, 1, 0, 1]]})
    gt_test = GT(df)

    with pytest.warns(
        UserWarning,
        match="Spacing is too large relative to the width. No bars will be displayed.",
    ):
        gt_plt_winloss(
            gt=gt_test,
            column="games",
            width=10,
            spacing=5,
        )


def test_gt_plt_bar_stack_snap(snapshot):
    df = pd.DataFrame({"team": ["A", "B"], "values": [[10, 20], [40, 30]]})
    gt_test = GT(df)
    res = gt_plt_bar_stack(
        gt=gt_test, column="values", palette=["red", "blue", "green"]
    )

    assert_rendered_body(snapshot, gt=res)


def test_gt_plt_bar_stack_basic():
    df = pd.DataFrame({"team": ["A", "B"], "values": [[10, 20, 30], [40, 30, 20]]})
    gt_test = GT(df)

    result = gt_plt_bar_stack(gt=gt_test, column="values")
    html = result.as_raw_html()

    assert html.count('width="32.0"') == 2
    assert html.count('height="30"') == 8
    assert html.count("<rect") == 6

    assert html.count('height="30" fill="#000000"') == 2
    assert html.count('height="30" fill="#25bce6"') == 2
    assert html.count('height="30" fill="#9e9e9e"') == 2

    assert html.count('<text dominant-baseline="central" text-anchor="middle"') == 6


def test_gt_plt_bar_stack_custom_palette():
    df = pd.DataFrame({"team": ["A"], "values": [[10, 20, 30]]})
    gt_test = GT(df)

    result = gt_plt_bar_stack(
        gt=gt_test,
        column="values",
        palette=["red", "blue", "green"],
    )
    html = result.as_raw_html()

    assert 'fill="red"' in html
    assert 'fill="blue"' in html
    assert 'fill="green"' in html


def test_gt_plt_bar_stack_custom_dimensions():
    df = pd.DataFrame({"team": ["A"], "values": [[10, 20, 30]]})
    gt_test = GT(df)

    result = gt_plt_bar_stack(
        gt=gt_test,
        column="values",
        width=200,
        height=50,
    )
    html = result.as_raw_html()

    assert '<svg xmlns="http://www.w3.org/2000/svg" width="200" height="50">' in html


def test_gt_plt_bar_stack_with_labels():
    df = pd.DataFrame({"team": ["A"], "values": [[10, 20, 30]]})
    gt_test = GT(df)

    result = gt_plt_bar_stack(
        gt=gt_test,
        column="values",
        labels=["Group 1", "Group 2", "Group 3"],
    )
    html = result.as_raw_html()

    assert "Group 1" in html
    assert "Group 2" in html
    assert "Group 3" in html


def test_gt_plt_bar_stack_relative_scaling():
    df = pd.DataFrame({"team": ["A", "B"], "values": [[1, 1], [2, 2]]})
    gt_test = GT(df)

    result = gt_plt_bar_stack(
        gt=gt_test,
        column="values",
        scale_type="relative",
    )
    html = result.as_raw_html()

    assert html.count('width="49.0"') == 4


def test_gt_plt_bar_stack_absolute_scaling():
    df = pd.DataFrame({"team": ["A", "B"], "values": [[1, 1], [2, 2]]})
    gt_test = GT(df)

    result = gt_plt_bar_stack(
        gt=gt_test,
        column="values",
        scale_type="absolute",
    )
    html = result.as_raw_html()

    assert html.count('width="49.0"') == 2
    assert html.count('width="24.5"') == 2


def test_gt_plt_bar_stack_with_empty_list():
    df = pd.DataFrame({"team": ["A", "B"], "values": [[], [10, 20, 30]]})
    gt_test = GT(df)

    result = gt_plt_bar_stack(gt=gt_test, column="values")
    html = result.as_raw_html()

    assert '<div style="width:100px; height:30px;"></div>' in html


def test_gt_plt_bar_stack_with_none_values():
    df = pd.DataFrame(
        {
            "team": ["A", "B", "C"],
            "values": [[None, 10, 20], [30, None, 40], [None, None, None]],
        }
    )
    gt_test = GT(df)
    result = gt_plt_bar_stack(gt=gt_test, column="values")
    html = result.as_raw_html()

    assert '<svg xmlns="http://www.w3.org/2000/svg" width="100" height="30"></svg>'
    assert html.count("<rect") == 4


def test_gt_plt_bar_stack_with_na_values():
    df = pd.DataFrame(
        {
            "team": ["A", "B", "C"],
            "values": [[np.nan, 10, 20], [30, np.nan, 40], [None, np.nan, None]],
        }
    )
    gt_test = GT(df)
    result = gt_plt_bar_stack(gt=gt_test, column="values")
    html = result.as_raw_html()

    assert '<svg xmlns="http://www.w3.org/2000/svg" width="100" height="30"></svg>'
    assert html.count("<rect") == 4


def test_gt_plt_bar_stack_spacing_warning():
    df = pd.DataFrame({"team": ["A"], "values": [[10, 20, 30]]})
    gt_test = GT(df)

    with pytest.warns(
        UserWarning,
        match="Spacing is too large relative to the width. No bars will be displayed.",
    ):
        result = gt_plt_bar_stack(
            gt=gt_test,
            column="values",
            width=10,
            spacing=5,
        )
        html = result.as_raw_html()

    assert (
        '<div style="display: flex;"><div style="width:10px; height:30px;"></div></div>'
        in html
    )


def test_gt_plt_bar_stack_invalid_column():
    df = pd.DataFrame({"team": ["A"], "values": [[10, 20, 30]]})
    gt_test = GT(df)

    with pytest.raises(KeyError):
        gt_plt_bar_stack(gt=gt_test, column="invalid_column")


def test_gt_plt_bar_stack_invalid_scale():
    df = pd.DataFrame({"team": ["A"], "values": [[10, 20, 30]]})
    gt_test = GT(df)

    with pytest.raises(ValueError):
        gt_plt_bar_stack(gt=gt_test, column="values", scale_type="invalid")  # type: ignore


def test_gt_plt_bar_pct_snap(snapshot, mini_gt):
    res = gt_plt_bar_pct(gt=mini_gt, column="num")

    assert_rendered_body(snapshot, gt=res)


def test_gt_plt_bar_pct(mini_gt):
    result = gt_plt_bar_pct(gt=mini_gt, column="num")
    html = result.as_raw_html()
    assert html.count("<svg") == 3


def test_gt_plt_bar_pct_autoscale_on(mini_gt):
    result = gt_plt_bar_pct(mini_gt, column="num", autoscale=True, labels=True)
    html = result.as_raw_html()
    assert ">100%</text>" in html


def test_gt_plt_bar_pct_autoscale_off(mini_gt):
    result = gt_plt_bar_pct(mini_gt, column="num", autoscale=False, labels=True)
    html = result.as_raw_html()
    assert ">33.3%</text>" in html


def test_gt_plt_bar_pct_without_labels(mini_gt):
    result = gt_plt_bar_pct(mini_gt, column="num", labels=False)
    html = result.as_raw_html()
    assert "</text>" not in html


def test_gt_plt_bar_pct_column_decimal(mini_gt):
    result = gt_plt_bar_pct(
        mini_gt, column="num", autoscale=False, labels=True, decimals=2
    )
    html = result.as_raw_html()
    assert ">33.33%</text>" in html


def test_gt_plt_bar_pct_label_placement():
    df = pd.DataFrame({"x": [10, 20, 30, 40]})
    gt = GT(df)
    result_autoscale_on = gt_plt_bar_pct(gt, "x", autoscale=True, labels=True)
    html_autoscale_on = result_autoscale_on.as_raw_html()

    assert html_autoscale_on.count('x="5.0px" y="8.0px"') == 3
    assert 'x="30.0px" y="8.0px"' in html_autoscale_on

    result_autoscale_off = gt_plt_bar_pct(gt, "x", autoscale=False, labels=True)
    html_autoscale_off = result_autoscale_off.as_raw_html()

    assert 'x="5.0px" y="8.0px"' in html_autoscale_off
    assert 'x="15.0px" y="8.0px"' in html_autoscale_off
    assert 'x="25.0px" y="8.0px"' in html_autoscale_off
    assert 'x="35.0px" y="8.0px"' in html_autoscale_off


def test_gt_plt_bar_pct_column_containing_effective_int():
    df = pd.DataFrame({"num": [1, 2.0]})
    result = gt_plt_bar_pct(GT(df), column="num", autoscale=False, labels=True)
    html = result.as_raw_html()
    assert ">1%</text>" in html
    assert ">2%</text>" in html


@pytest.mark.parametrize("height, width", [(16, 100), (17, 101)])
def test_gt_plt_bar_pct_height_width(mini_gt, height, width):
    result = gt_plt_bar_pct(
        mini_gt, column="num", height=height, width=width, labels=True
    )
    html = result.as_raw_html()
    assert f'height="{height}px">' in html
    assert f'width="{width}px"' in html


@pytest.mark.parametrize(
    "font_style, font_size", [("bold", 10), ("italic", 11), ("normal", 12)]
)
def test_gt_plt_bar_pct_fnot_style_size(mini_gt, font_style, font_size):
    result = gt_plt_bar_pct(
        mini_gt,
        column="num",
        labels=True,
        font_style=font_style,
        font_size=font_size,
    )
    html = result.as_raw_html()
    assert f'font-style="{font_style}"' in html
    assert f'font-size="{font_size}px"' in html


def test_gt_plt_bar_pct_column_containing_some_none():
    df = pd.DataFrame({"num": [1, None, None]})
    result = gt_plt_bar_pct(GT(df), column="num")
    html = result.as_raw_html()
    assert html.count('fill="transparent"/>') == 2


def test_gt_plt_bar_pct_column_containing_all_none():
    df = pd.DataFrame({"num": [None, None, None]})
    with pytest.raises(ValueError, match="All values in the column are None."):
        gt_plt_bar_pct(GT(df), column="num")


def test_gt_plt_bar_pct_label_cutoff_invalid_number(mini_gt):
    with pytest.raises(
        ValueError, match="Label_cutoff must be a number between 0 and 1."
    ):
        gt_plt_bar_pct(mini_gt, column="num", label_cutoff=100)


def test_gt_plt_bar_pct_font_style_invalid_string(mini_gt):
    with pytest.raises(
        ValueError, match="Font_style must be one of 'bold', 'italic', or 'normal'."
    ):
        gt_plt_bar_pct(mini_gt, column="num", font_style="invalid")  # type: ignore


def test_gt_plt_bullet_snap(snapshot):
    df = pd.DataFrame(
        {"name": ["A", "B", "C"], "actual": [10, 15, 25], "target": [12, 18, 20]}
    )
    gt_test = GT(df)
    res = gt_plt_bullet(gt=gt_test, data_column="actual", target_column="target")

    assert_rendered_body(snapshot, gt=res)


def test_gt_plt_bullet_basic():
    df = pd.DataFrame(
        {"name": ["A", "B", "C"], "actual": [10, 15, 25], "target": [12, 18, 20]}
    )
    gt_test = GT(df)
    result = gt_plt_bullet(gt=gt_test, data_column="actual", target_column="target")
    html = result.as_raw_html()

    assert html.count("<svg") == 3
    assert "target" not in html
    assert "actual" in html


def test_gt_plt_bullet_bar_height_too_high():
    df = pd.DataFrame({"actual": [10], "target": [12]})
    gt_test = GT(df)

    with pytest.warns(
        UserWarning,
        match="Bar_height must be less than or equal to the plot height. Adjusting bar_height to 100.",
    ):
        result = gt_plt_bullet(
            gt=gt_test,
            data_column="actual",
            target_column="target",
            bar_height=1000,
            height=100,
        )
        html = result.as_raw_html()

    assert html.count('height="100"') == 1
    assert 'height="1000"' not in html


def test_gt_plt_bullet_bar_height_too_low():
    df = pd.DataFrame({"actual": [10], "target": [12]})
    gt_test = GT(df)

    with pytest.warns(
        UserWarning,
        match="Bar_height cannot be negative. Adjusting bar_height to 0.",
    ):
        result = gt_plt_bullet(
            gt=gt_test,
            data_column="actual",
            target_column="target",
            bar_height=-100,
            height=1000,
        )
        html = result.as_raw_html()

    assert html.count('height="1000"') == 1
    assert 'height="-100"' not in html


def test_gt_plt_bullet_custom_colors():
    df = pd.DataFrame({"actual": [10, 20], "target": [15, 25]})
    gt_test = GT(df)

    result = gt_plt_bullet(
        gt=gt_test,
        data_column="actual",
        target_column="target",
        fill="blue",
        target_color="red",
    )
    html = result.as_raw_html()

    assert html.count('fill="blue"') == 2
    assert html.count('stroke="red"') == 2


def test_gt_plt_bullet_no_stroke_color():
    df = pd.DataFrame({"actual": [10], "target": [12]})
    gt_test = GT(df)

    result = gt_plt_bullet(
        gt=gt_test, data_column="actual", target_column="target", stroke_color=None
    )
    html = result.as_raw_html()

    assert 'stroke="transparent"' in html


def test_gt_plt_bullet_keep_data_column():
    df = pd.DataFrame({"actual": [10, 15], "target": [12, 18]})
    gt_test = GT(df)

    result = gt_plt_bullet(
        gt=gt_test, data_column="actual", target_column="target", keep_data_column=True
    )
    html = result.as_raw_html()

    assert ">actual plot</th>" in html
    assert ">actual</th>" in html
    assert html.count("<svg") == 2


def test_gt_plt_bullet_custom_dimensions():
    df = pd.DataFrame({"actual": [10, 18], "target": [12, 15]})
    gt_test = GT(df)

    result = gt_plt_bullet(
        gt=gt_test, data_column="actual", target_column="target", width=200, height=50
    )
    html = result.as_raw_html()

    assert html.count('width="200"') == 2
    assert html.count('height="50"') == 2


@pytest.mark.parametrize(
    "actual, target",
    [
        ([10, None, 30], [15, 25, None]),
        ([10, np.nan, 30], [15, 25, np.nan]),
    ],
)
def test_gt_plt_bullet_with_none_and_nan_values(actual, target):
    df = pd.DataFrame({"actual": actual, "target": target})
    gt_test = GT(df)

    result = gt_plt_bullet(gt=gt_test, data_column="actual", target_column="target")
    html = result.as_raw_html()

    assert isinstance(result, GT)
    assert html.count('<line stroke="darkgrey"') == 2
    assert html.count('fill="purple"') == 3
    assert html.count('width="0px" height="20px" fill="purple"') == 1


def test_gt_plt_bullet_invalid_data_column():
    df = pd.DataFrame({"actual": [10], "target": [12]})
    gt_test = GT(df)

    with pytest.raises(KeyError, match="Column 'invalid_col' not found"):
        gt_plt_bullet(gt=gt_test, data_column="invalid_col", target_column="target")


def test_gt_plt_bullet_invalid_target_column():
    df = pd.DataFrame({"actual": [10], "target": [12]})
    gt_test = GT(df)

    with pytest.raises(KeyError, match="Column 'invalid_col' not found"):
        gt_plt_bullet(gt=gt_test, data_column="actual", target_column="invalid_col")


def test_gt_plt_bullet_multiple_data_cols():
    df = pd.DataFrame({"actual": [10], "target": [12], "other": [5]})
    gt_test = GT(df)

    with pytest.raises(
        ValueError, match="Expected a single column, but got multiple columns"
    ):
        gt_plt_bullet(
            gt=gt_test, data_column=["actual", "other"], target_column="target"
        )


def test_gt_plt_bullet_multiple_target_cols():
    df = pd.DataFrame({"actual": [10], "target": [12], "other": [5]})
    gt_test = GT(df)

    with pytest.raises(
        ValueError, match="Expected a single column, but got multiple columns"
    ):
        gt_plt_bullet(
            gt=gt_test, data_column="actual", target_column=["target", "other"]
        )


def test_gt_plt_bullet_non_numeric_data_col():
    df = pd.DataFrame({"actual": ["text"], "target": [12]})
    gt_test = GT(df)

    with pytest.raises(TypeError, match="Invalid column type provided"):
        gt_plt_bullet(gt=gt_test, data_column="actual", target_column="target")


def test_gt_plt_bullet_non_numeric_target_col():
    df = pd.DataFrame({"actual": [10], "target": ["text"]})
    gt_test = GT(df)

    with pytest.raises(TypeError, match="Invalid column type provided"):
        gt_plt_bullet(gt=gt_test, data_column="actual", target_column="target")


def test_gt_plt_bullet_scaling():
    df = pd.DataFrame(
        {
            "actual": [10, 20, 30],
            "target": [40, 15, 25],
        }
    )
    gt_test = GT(df)

    result = gt_plt_bullet(gt=gt_test, data_column="actual", target_column="target")
    html = result.as_raw_html()

    assert html.count("<svg") == 3
    assert 'x1="58.5px" y1="0" x2="58.5px" y2="30px"' in html
    assert 'x1="21.0px" y1="0" x2="21.0px" y2="30px"' in html
    assert 'x1="36.0px" y1="0" x2="36.0px" y2="30px"' in html


def test_gt_plt_donut_snap(snapshot, mini_gt):
    res = gt_plt_donut(gt=mini_gt, columns="num")

    assert_rendered_body(snapshot, gt=res)


def test_gt_plt_donut_basic(mini_gt):
    result = gt_plt_donut(gt=mini_gt, columns=["num"])
    html = result.as_raw_html()

    assert html.count("<svg") == 3
    assert html.count('<svg style="fill-rule:true;"') == 1


def test_gt_plt_donut_show_labels_true(mini_gt):
    result = gt_plt_donut(gt=mini_gt, columns=["num"], show_labels=True)
    html = result.as_raw_html()

    assert ">33.33</text>" in html
    assert ">2.222</text>" in html
    assert ">0.1111</text>" in html
    assert html.count('<text dominant-baseline="central" text-anchor="middle"') == 3


def test_gt_plt_donut_show_labels_false(mini_gt):
    result = gt_plt_donut(gt=mini_gt, columns=["num"], show_labels=False)
    html = result.as_raw_html()

    assert "</text>" not in html


def test_gt_plt_donut_zero_values():
    df = pd.DataFrame({"values": [0, 10, 0]})
    gt_test = GT(df)
    result = gt_plt_donut(gt=gt_test, columns="values", show_labels=True)
    html = result.as_raw_html()

    assert ">0</text>" in html
    assert html.count('stroke-dasharray="3"') == 2


def test_gt_plt_donut_full_circle():
    df = pd.DataFrame({"values": [100, 100]})
    gt_test = GT(df)
    result = gt_plt_donut(gt=gt_test, columns="values")
    html = result.as_raw_html()

    assert 'style="fill-rule:true;"' in html


def test_gt_plt_donut_custom_colors(mini_gt):
    result = gt_plt_donut(
        gt=mini_gt,
        columns=["num"],
        fill="red",
        stroke_color="blue",
        label_color="green",
        show_labels=True,
    )
    html = result.as_raw_html()

    assert 'fill="red"' in html
    assert 'stroke="blue"' in html
    assert 'fill="green"' in html


def test_gt_plt_donut_no_stroke_color(mini_gt):
    result = gt_plt_donut(gt=mini_gt, columns=["num"], stroke_color=None)
    html = result.as_raw_html()
    assert 'stroke="transparent"' in html


def test_gt_plt_donut_keep_columns(mini_gt):
    result = gt_plt_donut(gt=mini_gt, columns=["num"], keep_columns=True)
    html = result.as_raw_html()

    assert ">num plot</th>" in html
    assert ">num</th>" in html
    assert html.count("<svg") == 3


def test_gt_plt_donut_with_na_values():
    df = pd.DataFrame({"values": [10, np.nan, 20, None]})
    gt_test = GT(df)
    result = gt_plt_donut(gt=gt_test, columns="values")
    html = result.as_raw_html()

    assert isinstance(result, GT)
    assert html.count('<div style="width:30px; height:30px;"></div>') == 2


def test_gt_plt_donut_custom_size(mini_gt):
    result = gt_plt_donut(gt=mini_gt, columns=["num"], size=50)
    html = result.as_raw_html()

    assert html.count('width="50" height="50"') == 3


def test_gt_plt_donut_with_domain_expanded(mini_gt):
    result = gt_plt_donut(gt=mini_gt, columns=["num"], domain=[0.1111, 33.33])
    html = result.as_raw_html()

    assert isinstance(result, GT)

    assert '<circle stroke="black" stroke-dasharray="3"' in html
    assert 'fill="transparent"' in html
    assert '<svg style="fill-rule:true;"' in html
    assert html.count("<svg") == 3


def test_gt_plt_donut_non_numeric_column(mini_gt):
    with pytest.raises(TypeError, match="Invalid column type provided"):
        gt_plt_donut(gt=mini_gt, columns="char")



from __future__ import annotations

from typing import Literal

from great_tables import GT, loc, style
from great_tables._locations import Loc
from great_tables._tbl_data import SelectExpr

__all__ = ["gt_add_divider"]


def gt_add_divider(
    gt: GT,
    columns: SelectExpr,
    sides: (
        Literal["right", "left", "top", "bottom", "all"]
        | list[Literal["right", "left", "top", "bottom", "all"]]
    ) = "right",
    color: str = "grey",
    divider_style: Literal["solid", "dashed", "dotted", "hidden", "double"] = "solid",
    weight: int = 2,
    include_labels: bool = True,
) -> GT:
    # TODO: include a simpler example first
    """
    Add dividers to specified columns in a `GT` object.

    The `gt_add_divider()` function takes an existing `GT` object and adds dividers to the specified
    columns. Dividers can be applied to one or more sides of the cells, with customizable color,
    style, and weight. Optionally, dividers can also be applied to column labels.

    Parameters
    ----------
    gt
        A `GT` object to modify.

    columns
        The columns to which dividers should be applied.

    sides
        The sides of the cells where dividers should be added. Options include `"right"`, `"left"`,
        `"top"`, `"bottom"`, or `"all"`. A list of sides can also be provided to apply dividers to
        multiple sides.

    color
        The color of the dividers.

    divider_style
        The style of the dividers. Options include `"solid"`, `"dashed"`, `"dotted"`, `"hidden"`,
        and `"double"`.

    weight
        The thickness of the dividers in pixels.

    include_labels
        Whether to include dividers in the column labels. If `True`, dividers will be applied to
        both the body and the column labels. If `False`, dividers will only be applied to the body.

    Returns
    -------
    GT
        A `GT` object with dividers added to the specified columns.

    Examples
    --------
    ```{python}
    import pandas as pd
    from great_tables import GT
    from great_tables.data import peeps
    import gt_extras as gte

    peeps_mini = peeps.head(6)

    gt = (
        GT(peeps_mini)
        .cols_hide([
            "name_family", "postcode", "country", "country_code",
            "dob", "gender", "state_prov", "email_addr",
        ])
        .tab_spanner("Location", ["address", "city"])
        .tab_spanner("Body Measurements", ["height_cm", "weight_kg"])
    )

    gt.pipe(
        gte.gt_add_divider,
        columns="name_given",
        color="#FFB90F",
        divider_style="double",
        weight=8,
    ).pipe(
        gte.gt_add_divider,
        columns="phone_number",
        color="purple",
        sides=["right", "left"],
        weight=5,
    )
    ```
    """

    locations: list[Loc] = [loc.body(columns=columns)]

    if include_labels:
        locations.append(loc.column_labels(columns=columns))

    res = gt.tab_style(
        style=style.borders(
            sides=sides,
            color=color,
            weight=f"{weight}px",
            style=divider_style,
        ),
        locations=locations,
    )

    return res



from __future__ import annotations

from great_tables import GT, google_font, loc, style

__all__ = [
    "gt_theme_538",
    "gt_theme_espn",
    "gt_theme_guardian",
    "gt_theme_nytimes",
    "gt_theme_excel",
    "gt_theme_dot_matrix",
    "gt_theme_dark",
    "gt_theme_pff",
]


def gt_theme_538(gt: GT) -> GT:
    """Applies a FiveThirtyEight-inspired theme to a `GT` object.

    This function styles a `GT` object with a look inspired by the FiveThirtyEight (538) website.

    Parameters
    ----------
    gt
        An existing `GT` object.

    Returns
    ----------
    GT
        The themed `GT` object, allowing for method chaining.

    Examples
    ----------
    ```{python}
    from great_tables import GT, md
    from great_tables.data import gtcars
    import gt_extras as gte

    gtcars_mini = gtcars[["model", "year", "hp", "trq"]].head(5)

    gt = (
        GT(gtcars_mini, rowname_col="model")
        .tab_stubhead(label=md("*Car*"))
        .tab_header(title="Car Collection")
    )

    gt.pipe(gte.gt_theme_538)
    ```
    """
    gt_themed = (
        gt.opt_table_font(font=google_font("Cairo"))
        .tab_style(
            style=style.text(font=google_font("Chivo"), weight="bold"),
            locations=loc.title(),
        )
        .tab_style(
            style=style.text(font=google_font("Chivo")),
            locations=loc.subtitle(),
        )
        .tab_style(
            style=[
                style.borders(sides="top", color="black", weight="0px"),
                style.text(
                    font=google_font("Chivo"),
                    transform="uppercase",
                    v_align="bottom",
                    size="14px",
                ),
            ],
            locations=[loc.column_labels(), loc.stubhead()],
        )
        # Altered wrt R package
        # .tab_style(
        #     style=style.borders(sides="bottom", color="black", weight="1px"),
        #     locations=loc.row_groups(),
        # )
        .tab_options(
            column_labels_background_color="white",
            data_row_padding="3px",
            heading_border_bottom_style="none",
            table_border_top_width="3px",
            table_border_top_style="none",
            table_border_bottom_style="none",
            column_labels_font_weight="normal",
            column_labels_border_top_style="none",
            column_labels_border_bottom_width="2px",
            column_labels_border_bottom_color="black",
            row_group_border_top_style="none",
            row_group_border_top_color="black",
            row_group_border_bottom_width="1px",
            # row_group_border_bottom_color="white", # Altered wrt R package
            row_group_border_bottom_color="black",  # Altered wrt R package
            stub_border_color="white",
            stub_border_width="0px",
            source_notes_font_size="12px",
            source_notes_border_lr_style="none",
            table_font_size="16px",
            heading_align="left",
        )
    )

    return gt_themed


def gt_theme_espn(gt: GT) -> GT:
    """Applies an ESPN-inspired theme to a `GT` object.

    This function styles a `GT` object with a look inspired by ESPN's data tables.

    Parameters
    ----------
    gt
        An existing `GT` object.

    Returns
    -------
    GT
        The themed `GT` object, allowing for method chaining.

    Examples
    ----------
    ```{python}
    from great_tables import GT, md
    from great_tables.data import gtcars
    import gt_extras as gte

    gtcars_mini = gtcars[["model", "year", "hp", "trq"]].head(5)

    gt = (
        GT(gtcars_mini, rowname_col="model")
        .tab_stubhead(label=md("*Car*"))
        .tab_header(title="Car Collection")
    )

    gt.pipe(gte.gt_theme_espn)
    ```
    """
    gt_themed = (
        gt.opt_all_caps()
        .opt_table_font(font=google_font("Lato"), weight=400)
        .opt_row_striping()
        .tab_style(style=style.text(weight="bold"), locations=loc.column_header())
        .tab_options(
            row_striping_background_color="#fafafa",
            table_body_hlines_color="#f6f7f7",
            source_notes_font_size="12px",
            table_font_size="16px",
            heading_align="left",
            heading_title_font_size="24px",
            table_border_top_color="white",
            table_border_top_width="3px",
            data_row_padding="7px",
        )
    )

    return gt_themed


def gt_theme_nytimes(gt: GT) -> GT:
    """Applies a New York Times-inspired theme to a `GT` object.

    This function styles a `GT` object with a look inspired by New York Times tables.

    Parameters
    ----------
    gt
        An existing `GT` object.

    Returns
    -------
    GT
        The themed `GT` object, allowing for method chaining.

    Examples
    ----------
    ```{python}
    from great_tables import GT, md
    from great_tables.data import gtcars
    import gt_extras as gte

    gtcars_mini = gtcars[["model", "year", "hp", "trq"]].head(5)

    gt = (
        GT(gtcars_mini, rowname_col="model")
        .tab_stubhead(label=md("*Car*"))
        .tab_header(title="Car Collection")
    )

    gt.pipe(gte.gt_theme_nytimes)
    ```
    """
    gt_themed = (
        gt.tab_style(
            style=style.text(
                color="darkgrey",
                font=google_font("Source Sans Pro"),
                transform="uppercase",
            ),
            locations=[loc.column_labels(), loc.stubhead()],
        )
        .tab_style(
            style=style.text(font=google_font("Libre Franklin"), weight="bolder"),
            locations=loc.title(),
        )
        .tab_style(
            style=style.text(font=google_font("Source Sans Pro")),
            locations=loc.body(),
        )
        .tab_options(
            heading_align="left",
            column_labels_border_top_style="none",
            table_border_top_style="none",
            column_labels_border_bottom_style="none",
            column_labels_border_bottom_width="1px",
            column_labels_border_bottom_color="#334422",
            table_body_border_top_style="none",
            table_body_border_bottom_color="white",
            heading_border_bottom_style="none",
            data_row_padding="7px",
            column_labels_font_size="12px",
        )
    )
    return gt_themed


def gt_theme_guardian(gt: GT) -> GT:
    """Applies a Guardian-inspired theme to a `GT` object.

    This function styles a `GT` object with a look inspired by The Guardian's data tables.

    Parameters
    ----------
    gt
        An existing `GT` object.

    Returns
    -------
    GT
        The themed `GT` object, allowing for method chaining.

    Examples
    ----------
    ```{python}
    from great_tables import GT, md
    from great_tables.data import gtcars
    import gt_extras as gte

    gtcars_mini = gtcars[["model", "year", "hp", "trq"]].head(5)

    gt = (
        GT(gtcars_mini, rowname_col="model")
        .tab_stubhead(label=md("*Car*"))
        .tab_header(title="Car Collection")
    )

    gt.pipe(gte.gt_theme_guardian)
    ```
    """
    ## Altered wrt R package to not include whitespace between lines
    gt_themed = (
        gt.opt_table_font(font=google_font("Noto Sans"))
        ## Altered wrt R package
        # .tab_style(
        #     ## style hidden or weight 0px?
        #     style=style.borders(sides="top", color="white", style="hidden"),
        #     ## A place we might see a difference from R
        #     locations=loc.body(rows=0),
        # )
        .tab_style(
            style=style.text(color="#005689", size="22px", weight="bold"),
            locations=loc.title(),
        )
        .tab_style(
            style=style.text(color="#005689", size="16px", weight="bold"),
            locations=loc.subtitle(),
        )
        .tab_options(
            row_striping_include_table_body=True,
            table_background_color="#f6f6f6",
            row_striping_background_color="#ececec",
            column_labels_background_color="#f6f6f6",
            column_labels_font_weight="bold",
            table_border_top_width="1px",
            table_border_top_color="#40c5ff",
            table_border_bottom_width="3px",
            table_border_bottom_color="white",
            source_notes_border_bottom_width="0px",
            table_body_border_bottom_width="3px",
            table_body_border_bottom_color="white",
            table_body_hlines_width="0px",  # Raise issue in gtExtras R
            table_body_hlines_color="white",
            row_group_border_top_width="1px",
            row_group_border_top_color="grey",
            row_group_border_bottom_width="1px",
            row_group_border_bottom_color="grey",
            row_group_font_weight="bold",
            column_labels_border_top_width="1px",
            # Slight modification from the R version:
            column_labels_border_top_color="#ececec"
            if gt._heading.title
            else "#40c5ff",
            column_labels_border_bottom_width="2px",
            column_labels_border_bottom_color="#ececec",
            heading_border_bottom_width="0px",
            data_row_padding="4px",
            source_notes_font_size="12px",
            table_font_size="16px",
            heading_align="left",
        )
        # this replaces footnotes_border_bottom_width="0px", because that functionality doesn't
        # exist in the Python API
        .tab_style(
            style=style.borders(sides="bottom", style="hidden"), locations=loc.footer()
        )
    )
    return gt_themed


def gt_theme_excel(gt: GT, color: str = "lightgrey") -> GT:
    """Applies an Excel-inspired theme to a `GT` object.

    This function styles a `GT` object with a look inspired by Microsoft Excel tables.

    Parameters
    ----------
    gt
        An existing `GT` object.

    color
        A string indicating the color of the row striping, defaults to a light grey.
        Accepts either named colors or hex colors.

    Returns
    -------
    GT
        The themed `GT` object, allowing for method chaining.

    Examples
    ----------
    ```{python}
    from great_tables import GT, md
    from great_tables.data import gtcars
    import gt_extras as gte

    gtcars_mini = gtcars[["model", "year", "hp", "trq"]].head(5)

    gt = (
        GT(gtcars_mini, rowname_col="model")
        .tab_stubhead(label=md("*Car*"))
        .tab_header(title="Car Collection")
    )

    gt.pipe(gte.gt_theme_excel)
    ```
    """
    gt_themed = (
        gt.opt_row_striping()
        .tab_style(
            style=style.borders(sides="all", weight="1px", color="black"),
            locations=loc.body(),
        )
        # This does not appear to achieve anything
        # .tab_style(
        #     style=style.borders(sides="left", weight="2px", color="black"),
        #     locations=[
        #         loc.body(columns=0),
        #         loc.column_labels(columns=0),
        #         loc.stub()
        #     ],
        # )
        # This does not appear to achieve anything
        # .tab_style(
        #     style=style.borders(sides="left", weight="1px", color="black"),
        #     locations=loc.row_groups(),
        # )
        .opt_table_font(font="Calibri")
        .tab_options(
            heading_align="left",
            heading_border_bottom_color="black",
            column_labels_background_color="black",
            column_labels_font_weight="bold",
            stub_background_color="white",
            stub_border_color="black",
            row_group_background_color="white",
            row_group_border_top_color="black",
            row_group_border_bottom_color="black",
            row_group_border_left_color="black",
            row_group_border_right_color="black",
            row_group_border_left_width="1px",
            row_group_border_right_width="1px",
            column_labels_font_size="85%",
            column_labels_border_top_style="none",
            column_labels_border_bottom_color="black",
            column_labels_border_bottom_width="2px",
            table_border_left_color="black",
            table_border_left_style="solid",
            table_border_right_style="solid",
            table_border_left_width="2px",
            table_border_right_width="2px",
            table_border_right_color="black",
            table_border_bottom_width="2px",
            table_border_bottom_color="black",
            table_border_top_width="2px",
            table_border_top_color="black",
            row_striping_background_color=color,
            table_body_hlines_color="black",
            table_body_vlines_color="black",
            data_row_padding="1px",
        )
    )
    return gt_themed


def gt_theme_dot_matrix(gt: GT, color: str = "#b5dbb6") -> GT:
    """Applies a dot-matrix-inspired theme to a `GT` object.

    This function styles a `GT` object with a look reminiscent of dot-matrix printouts.

    Parameters
    ----------
    gt
        An existing `GT` object.

    color
        A string indicating the color of the row striping, defaults to `"#b5dbb6"`.
        Accepts either named colors or hex colors.

    Returns
    -------
    GT
        The themed `GT` object, allowing for method chaining.

    Examples
    ----------
    ```{python}
    from great_tables import GT, md
    from great_tables.data import gtcars
    import gt_extras as gte

    gtcars_mini = gtcars[["model", "year", "hp", "trq"]].head(5)

    gt = (
        GT(gtcars_mini, rowname_col="model")
        .tab_stubhead(label=md("*Car*"))
        .tab_header(title="Car Collection")
    )

    gt.pipe(gte.gt_theme_dot_matrix)
    ```
    """
    gt_themed = (
        gt.opt_row_striping()
        .opt_table_font(font="Courier")
        .tab_options(
            heading_align="left",
            heading_border_bottom_color="white",
            column_labels_text_transform="lowercase",
            column_labels_font_size="85%",
            column_labels_border_top_style="none",
            column_labels_border_bottom_color="black",
            column_labels_border_bottom_width="2px",
            table_border_bottom_style="none",
            table_border_bottom_width="2px",
            table_border_bottom_color="white",
            table_border_top_style="none",
            row_striping_background_color=color,
            table_body_hlines_style="none",
            table_body_vlines_style="none",
            data_row_padding="1px",
        )
    )

    return gt_themed


def gt_theme_dark(gt: GT) -> GT:
    """Applies a dark mode theme to a `GT` object.

    This function styles a `GT` object with a dark background and light text.

    Parameters
    ----------
    gt
        An existing `GT` object.

    Returns
    -------
    GT
        The themed `GT` object, allowing for method chaining.

    Examples
    ----------
    ```{python}
    from great_tables import GT, md
    from great_tables.data import gtcars
    import gt_extras as gte

    gtcars_mini = gtcars[["model", "year", "hp", "trq"]].head(5)

    gt = (
        GT(gtcars_mini, rowname_col="model")
        .tab_stubhead(label=md("*Car*"))
        .tab_header(title="Car Collection")
    )

    gt.pipe(gte.gt_theme_dark)
    ```
    """
    gt_themed = (
        gt.tab_style(
            style=style.text(
                color="white",
                font=google_font("Source Sans Pro"),
                transform="uppercase",
            ),
            locations=[loc.column_labels(), loc.stubhead()],
        )
        .tab_style(
            style=style.text(font=google_font("Libre Franklin"), weight="bolder"),
            locations=loc.title(),
        )
        .tab_style(
            style=style.text(font=google_font("Source Sans Pro")),
            locations=loc.body(),
        )
        .tab_options(
            heading_align="left",
            heading_border_bottom_style="none",
            table_background_color="#333333",
            table_font_color_light="white",
            table_border_top_style="none",
            table_border_bottom_color="#333333",
            table_border_left_color="#333333",
            table_border_right_color="#333333",
            table_body_border_top_style="none",
            table_body_border_bottom_color="#333333",
            column_labels_border_top_style="none",
            column_labels_background_color="#333333",
            column_labels_border_bottom_width="3px",
            column_labels_border_bottom_color="white",
            data_row_padding="7px",
        )
    )
    return gt_themed


def gt_theme_pff(
    gt: GT,
    divider: str | None = None,
    spanners: list[str] | None = None,
    rank_col: str | None = None,
) -> GT:
    """Applies a Pro Football Focus (PFF)-inspired theme to a `GT` object.

    This function styles a `GT` object with a look inspired by Pro Football Focus tables,
    supporting custom spanners, dividers, and rank column highlighting.

    Parameters
    ----------
    gt
        An existing `GT` object.

    divider
        Zero or more column names/indices to visually divide with a preceding border.

    spanners
        Optional list of spanners to style, as referenced by the `GT` spanner ids.

    rank_col
        Optional single column name/index to highlight as a rank column.

    Returns
    -------
    GT
        The themed `GT` object, allowing for method chaining.

    Examples
    ----------
    ```{python}
    from great_tables import GT, md
    from great_tables.data import gtcars
    import gt_extras as gte

    gtcars_mini = gtcars[["model", "year", "hp", "trq"]].head(5)

    gt = (
        GT(gtcars_mini, rowname_col="model")
        .tab_stubhead(label=md("*Car*"))
        .tab_header(title="Car Collection")
    )

    gte.gt_theme_pff(gt, rank_col="trq")
    ```
    """
    gt_themed = (
        gt.opt_row_striping()
        .opt_all_caps()
        .tab_options(
            table_body_hlines_color="transparent",
            table_border_top_width="3px",
            table_border_top_color="transparent",
            table_border_bottom_color="lightgrey",
            table_border_bottom_width="1px",
            column_labels_border_top_width="3px",
            column_labels_padding="6px",
            column_labels_border_top_color="transparent",
            column_labels_border_bottom_width="3px",
            column_labels_border_bottom_color="transparent",
            row_striping_background_color="#f5f5f5",
            data_row_padding="6px",
            heading_align="left",
            heading_title_font_size="30px",
            heading_title_font_weight="bold",
            heading_subtitle_font_size="16px",
            table_font_size="12px",
        )
        .opt_table_font(font=google_font("Roboto"))
    )

    # Handle spanners if provided
    if spanners:
        span_cols = [col for spanner in gt._spanners for col in spanner.vars]

        # Add a blank spanner
        gt_themed = (
            gt_themed.tab_spanner(
                columns=[
                    str(col)
                    for col in gt._boxhead._get_column_labels()
                    if col not in span_cols
                ],
                label=" ",
                id="blank",
            )
            .tab_style(
                style=[
                    style.fill(color="transparent"),
                    style.text(color="transparent", size="9px", weight="bold"),
                    style.borders(sides="left", color="transparent", weight="3px"),
                    style.borders(sides="top", color="transparent", weight="3px"),
                ],
                locations=loc.spanner_labels(ids=["blank"]),
            )
            # Add real spanners and style
            .tab_style(
                style=[
                    style.fill(color="#f5f5f5"),
                    style.text(color="#878e94", size="10px", weight="bold"),
                    style.borders(sides="left", color="white", weight="3px"),
                    style.borders(sides="top", color="white", weight="3px"),
                ],
                locations=loc.spanner_labels(ids=spanners),
            )
        )

    # Handle divider if provided
    if divider:
        gt_themed = gt_themed.tab_style(
            style=style.borders(sides="left", color="lightgrey", weight="2px"),
            locations=loc.body(columns=divider),
        ).tab_style(
            style=style.borders(sides="left", color="#212426", weight="2px"),
            locations=loc.column_labels(columns=divider),
        )

    # Handle rank_col if provided
    if rank_col:
        gt_themed = gt_themed.tab_style(
            style=[
                style.fill(color="#e4e8ec"),
                style.borders(color="#e4e8ec"),
            ],
            locations=loc.body(columns=rank_col),
        ).cols_align("center", columns=rank_col)

    gt_themed = gt_themed.tab_style(
        style=[
            style.fill(color="#585d63"),
            style.text(color="white", size="10px", weight="bold"),
            style.borders(sides="bottom", color="#585d63", weight="2.5px"),
        ],
        locations=[loc.column_labels(), loc.stubhead()],
    )

    return gt_themed



from __future__ import annotations

import math
import warnings
from typing import TYPE_CHECKING, Literal

from great_tables import GT, html
from great_tables._data_color.base import (
    _html_color,
    _ideal_fgnd_color,
)
from great_tables._locations import resolve_cols_c
from great_tables._tbl_data import SelectExpr, is_na
from scipy.stats import sem, t, tmean
from svg import (
    SVG,
    Arc,
    Circle,
    ClosePath,
    Length,
    Line,
    LineTo,
    MoveTo,
    Path,
    Rect,
    Text,
)

from gt_extras import gt_duplicate_column
from gt_extras._utils_color import _get_discrete_colors_from_palette
from gt_extras._utils_column import (
    _format_numeric_text,
    _scale_numeric_column,
    _validate_and_get_single_column,
)

__all__ = [
    "gt_plt_bar",
    "gt_plt_bar_pct",
    "gt_plt_bar_stack",
    "gt_plt_bullet",
    "gt_plt_conf_int",
    "gt_plt_donut",
    "gt_plt_dot",
    "gt_plt_dumbbell",
    "gt_plt_winloss",
]

# TODO: default font for labels?

# TODO: how to handle negative values? Plots can't really have negative length


def gt_plt_bar(
    gt: GT,
    columns: SelectExpr = None,
    fill: str = "purple",
    bar_height: float = 20,
    height: float = 30,
    width: float = 60,
    stroke_color: str | None = "black",
    show_labels: bool = False,
    label_color: str = "white",
    domain: list[int] | list[float] | None = None,
    keep_columns: bool = False,
) -> GT:
    """
    Create horizontal bar plots in `GT` cells.

    The `gt_plt_bar()` function takes an existing `GT` object and adds horizontal bar charts to
    specified columns. Each cell value is represented as a horizontal bar with length proportional
    to the cell's numeric value relative to the column's maximum value.

    Parameters
    ----------
    gt
        A `GT` object to modify.

    columns
        The columns to target. Can be a single column or a list of columns (by name or index).
        If `None`, the bar plot is applied to all numeric columns.

    fill
        The fill color for the bars.

    bar_height
        The height of each individual bar in pixels.

    height
        The height of the bar plot in pixels. In practice, this allows for the bar to appear
        less stout, the larger the difference between `height` and `bar_height`.

    width
        The width of the maximum bar in pixels. Not all bars will have this width.

    stroke_color
        The color of the vertical axis on the left side of the bar. The default is black, but if
        `None` is passed, no stroke will be drawn.

    show_labels
        Whether or not to show labels on the bars.

    label_color
        The color of text labels on the bars (when `show_labels` is `True`).

    keep_columns
        Whether to keep the original column values. If this flag is `True`, the plotted values will
        be duplicated into a new column with the string " plot" appended to the end of the column
        name. See [`gt_duplicate_column()`](https://posit-dev.github.io/gt-extras/reference/gt_duplicate_column)
        for more details.


    Returns
    -------
    GT
        A `GT` object with horizontal bar plots added to the specified columns.

    Examples
    --------

    ```{python}
    from great_tables import GT
    from great_tables.data import gtcars
    import gt_extras as gte

    gtcars_mini = gtcars.loc[
        9:17,
        ["model", "mfr", "year", "hp", "hp_rpm", "trq", "trq_rpm", "mpg_c", "mpg_h"]
    ]

    gt = (
        GT(gtcars_mini, rowname_col="model")
        .tab_stubhead(label="Car")
        .cols_align("center")
        .cols_align("left", columns="mfr")
    )

    gt.pipe(
        gte.gt_plt_bar,
        columns= ["hp", "hp_rpm", "trq", "trq_rpm", "mpg_c", "mpg_h"]
    )
    ```

    Note
    --------
    Each column's bars are scaled independently based on that column's min/max values.
    """
    if bar_height > height:
        bar_height = height
        warnings.warn(
            f"Bar_height must be less than or equal to the plot height. Adjusting bar_height to {bar_height}.",
            category=UserWarning,
        )

    if bar_height < 0:
        bar_height = 0
        warnings.warn(
            f"Bar_height cannot be negative. Adjusting bar_height to {bar_height}.",
            category=UserWarning,
        )

    # Allow the user to hide the vertical stroke
    if stroke_color is None:
        stroke_color = "transparent"

    def _make_bar(scaled_val: float, original_val: int | float) -> str:
        svg = _make_bar_svg(
            scaled_val=scaled_val,
            original_val=original_val,
            fill=fill,
            bar_height=bar_height,
            height=height,
            width=width,
            stroke_color=stroke_color,
            show_labels=show_labels,
            label_color=label_color,
        )

        return f'<div style="display: flex;">{svg.as_str()}</div>'

    # Get names of columns
    columns_resolved = resolve_cols_c(data=gt, expr=columns)

    res = gt
    for column in columns_resolved:
        # Validate this is a single column and get values
        col_name, col_vals = _validate_and_get_single_column(
            gt,
            column,
        )

        scaled_vals = _scale_numeric_column(
            res._tbl_data,
            col_name,
            col_vals,
            domain,
        )

        # The location of the plot column will be right after the original column
        if keep_columns:
            res = gt_duplicate_column(
                res,
                col_name,
                after=col_name,
                append_text=" plot",
            )
            col_name = col_name + " plot"

        # Apply the scaled value for each row, so the bar is proportional
        for i, scaled_val in enumerate(scaled_vals):
            res = res.fmt(
                lambda original_val, scaled_val=scaled_val: _make_bar(
                    original_val=original_val,
                    scaled_val=scaled_val,
                ),
                columns=col_name,
                rows=[i],
            )

    return res


def gt_plt_bullet(
    gt: GT,
    data_column: SelectExpr,
    target_column: SelectExpr,
    # target_value: int | None = None, # if this, let target_column be none?
    fill: str = "purple",
    bar_height: float = 20,
    height: float = 30,
    width: float = 60,
    target_color: str = "darkgrey",
    stroke_color: str | None = "black",
    # show_labels: bool = False, # Maybe include in later version of fn, to label target or data?
    # label_color: str = "white",
    keep_data_column: bool = False,
) -> GT:
    """
    Create bullet chart plots in `GT` cells.

    The `gt_plt_bullet()` function takes an existing `GT` object and adds bullet chart
    visualizations to compare actual values against target values. Each bullet chart consists
    of a horizontal bar representing the actual value and a vertical line indicating the target
    value, making it easy to assess performance against goals or benchmarks.

    Parameters
    ----------
    gt
        A `GT` object to modify.

    data_column
        The column containing the actual values to be plotted as horizontal bars.

    target_column
        The column containing the target values to be displayed as vertical reference lines.
        This column will be automatically hidden from the returned table.

    fill
        The fill color for the horizontal bars representing actual values.

    bar_height
        The height of each horizontal bar in pixels.

    height
        The height of the bullet chart plot in pixels. This allows for spacing around
        the bar and target line.

    width
        The width of the maximum bar in pixels. Bars are scaled proportionally to this width.

    target_color
        The color of the vertical target line.

    stroke_color
        The color of the vertical axis on the left side of the chart. The default is black, but if
        `None` is passed, no stroke will be drawn.

    keep_data_column
        Whether to keep the original data column values. If this flag is `True`, the plotted values
        will be duplicated into a new column with the string " plot" appended to the end of the
        column name. See [`gt_duplicate_column()`](https://posit-dev.github.io/gt-extras/reference/gt_duplicate_column)
        for more details.

    Returns
    -------
    GT
        A `GT` object with bullet chart plots added to the specified data column. The target
        column is automatically hidden from the table.

    Examples
    --------
    ```{python}
    import polars as pl
    from great_tables import GT
    from great_tables.data import airquality
    import gt_extras as gte

    air_bullet = (
        pl.from_pandas(airquality)
        .with_columns(pl.col("Temp").mean().over("Month").alias("target_temp"))
        .group_by("Month", maintain_order=True)
        .head(2)
        .with_columns(
            (pl.col("Month").cast(pl.Utf8) + "/" + pl.col("Day").cast(pl.Utf8)).alias(
                "Date"
            )
        )
        .select(["Date", "Temp", "target_temp"])
        .with_columns(pl.col(["Temp", "target_temp"]).round(1))
    )

    (
        GT(air_bullet, rowname_col="Date")
        .tab_header(title="Daily Temp vs Monthly Average")
        .tab_source_note("Target line shows monthly average temperature")

        ## Call gt_plt_bullet
        .pipe(
            gte.gt_plt_bullet,
            data_column="Temp",
            target_column="target_temp",
            width=200,
            fill="tomato",
            target_color="darkblue",
            keep_data_column=True,
        )
        .cols_move_to_end("Temp")
        .cols_align("left", "Temp plot")
    )
    ```

    Note
    ----
    Both data and target values are scaled to a common domain for consistent visualization.
    The scaling domain is automatically determined as `[0, max(data_values, target_values)]`.
    """
    if bar_height > height:
        bar_height = height
        warnings.warn(
            f"Bar_height must be less than or equal to the plot height. Adjusting bar_height to {bar_height}.",
            category=UserWarning,
        )

    if bar_height < 0:
        bar_height = 0
        warnings.warn(
            f"Bar_height cannot be negative. Adjusting bar_height to {bar_height}.",
            category=UserWarning,
        )

    # Allow the user to hide the vertical stroke
    if stroke_color is None:
        stroke_color = "transparent"

    def _make_bullet_plot_svg(
        scaled_val: float,
        original_val: int | float,
        target_val: float,
        original_target_val: float | int | None,
    ) -> str:
        svg = _make_bar_svg(
            scaled_val=scaled_val,
            original_val=original_val,
            fill=fill,
            bar_height=bar_height,
            height=height,
            width=width,
            stroke_color=stroke_color,
            show_labels=False,
            label_color="black",  # placeholder
        )

        # this should never be reached, but is needed for the type checker
        if TYPE_CHECKING:
            if svg.elements is None:
                raise ValueError(
                    "Unreachable code: svg.elements should never be None here."
                )

        if not is_na(res._tbl_data, original_target_val):
            _stroke_width = height / 10
            _x_location = max(_stroke_width, width * target_val - _stroke_width / 2)

            svg.elements.append(
                Line(
                    x1=Length(_x_location, "px"),
                    x2=Length(_x_location, "px"),
                    y1=0,
                    y2=Length(height, "px"),
                    stroke_width=Length(_stroke_width, "px"),
                    stroke=target_color,
                ),
            )

        return f'<div style="display: flex;">{svg.as_str()}</div>'

    res = gt

    data_col_name, data_col_vals = _validate_and_get_single_column(
        gt,
        data_column,
    )
    target_col_name, target_col_vals = _validate_and_get_single_column(
        gt,
        target_column,
    )

    # Only consider numeric values for scaling domain
    domain = None
    numeric_data_vals = [v for v in data_col_vals if isinstance(v, (int, float))]
    numeric_target_vals = [v for v in target_col_vals if isinstance(v, (int, float))]
    if numeric_data_vals or numeric_target_vals:
        domain = [0, max([*numeric_data_vals, *numeric_target_vals])]

    scaled_data_vals = _scale_numeric_column(
        res._tbl_data,
        data_col_name,
        data_col_vals,
        domain,
    )

    scaled_target_vals = _scale_numeric_column(
        res._tbl_data,
        target_col_name,
        target_col_vals,
        domain,
    )

    if keep_data_column:
        res = gt_duplicate_column(
            res,
            data_col_name,
            after=data_col_name,
            append_text=" plot",
        )
        data_col_name = data_col_name + " plot"

    # Apply the scaled value for each row, so the bar is proportional
    for i, scaled_val in enumerate(scaled_data_vals):
        target_val = scaled_target_vals[i]
        original_target_val = target_col_vals[i]

        res = res.fmt(
            lambda original_val,
            scaled_val=scaled_val,
            target_val=target_val,
            original_target_val=original_target_val: _make_bullet_plot_svg(
                original_val=original_val,
                scaled_val=scaled_val,
                target_val=target_val,
                original_target_val=original_target_val,
            ),
            columns=data_col_name,
            rows=[i],
        )

    res = res.cols_hide(target_col_name)

    return res


def gt_plt_dot(
    gt: GT,
    category_col: SelectExpr,
    data_col: SelectExpr,
    width: int = 120,
    height: int = 30,
    font_size: int = 16,
    domain: list[int] | list[float] | None = None,
    palette: list[str] | str | None = None,
) -> GT:
    """
    Create dot plots with thin horizontal bars in `GT` cells.

    The `gt_plt_dot()` function takes an existing `GT` object and adds dot plots with horizontal
    bar charts to a specified category column. Each cell displays a colored dot according to the
    value in the assigned category column and a horizontal bar representing the corresponding
    numeric value from the data column.

    Parameters
    ----------
    gt
        A `GT` object to modify.

    category_col
        The column containing category labels that will be displayed next to colored dots. The
        coloring of the dots are determined by this column.

    data_col
        The column containing numeric values that will determine the length of the horizontal bars.

    width
        The width of the SVG plot in pixels. You may need to increase this if your category labels
        are long.

    height
        The height of the SVG plot in pixels.

    font_size
        The font size for the category label text.

    domain
        The domain of values to use for the color scheme. This can be a list of floats or integers.
        If `None`, the domain is automatically set to `[0, max(data_col)]`.

    palette
        The color palette to use. This should be a list of colors
        (e.g., `["#FF0000"`, `"#00FF00"`, `"#0000FF"` `]`). A ColorBrewer palette could also be used,
        just supply the name (see [`GT.data_color()`](https://posit-dev.github.io/great-tables/reference/GT.data_color) for additional reference).
        If `None`, then a default palette will be used.

    Returns
    -------
    GT
        A `GT` object with dot plots and horizontal bars added to the specified category column.

    Examples
    --------
    ```{python}
    from great_tables import GT
    from great_tables.data import gtcars
    import gt_extras as gte

    gtcars_mini = gtcars.loc[8:20, ["model", "mfr", "hp"]]

    gt = (
        GT(gtcars_mini, rowname_col="model")
        .tab_stubhead(label="Car")
    )

    gt.pipe(gte.gt_plt_dot, category_col="mfr", data_col="hp")
    ```

    Note
    -------
    If your category label text is cut off or does not fit, you likely need to increase the `width`
    parameter to allow more space for the text and plot.
    """
    # Get the underlying Dataframe
    data_table = gt._tbl_data

    def _make_dot_and_bar_svg(
        bar_val: float,
        fill: str,
        dot_category_label: str,
        svg_width: float,
        svg_height: float,
        font_size: float,
    ) -> str:
        if is_na(data_table, bar_val) or is_na(data_table, dot_category_label):
            return "<div></div>"

        # Layout parameters
        dot_radius = font_size / 2.75
        dot_x = dot_radius
        dot_y = svg_height / 2

        # Text positioning
        text_x = dot_x + dot_radius * 1.5
        text_y = dot_y

        # Bar positioning
        bar_y = text_y + (font_size / 2) * 1.2  # 1.2 for padding
        bar_height = svg_height / 8
        bar_start_x = 0
        bar_width = svg_width * bar_val

        elements = [
            # Dot
            Circle(
                cx=dot_x,
                cy=dot_y,
                r=dot_radius,
                fill=fill,
            ),
            # Category label text
            Text(
                text=dot_category_label,
                x=text_x,
                y=text_y,
                fill="black",
                font_size=font_size,
                dominant_baseline="central",
                text_anchor="start",
            ),
            # Bar
            Rect(
                x=bar_start_x,
                y=bar_y,
                width=bar_width,
                height=bar_height,
                fill=fill,
                rx=2,
            ),
        ]

        svg = SVG(width=svg_width, height=svg_height, elements=elements)
        return f'<div style="display: flex;">{svg.as_str()}</div>'

    # Validate and get data column
    data_col_name, data_col_vals = _validate_and_get_single_column(
        gt,
        data_col,
    )

    # Process numeric data column
    scaled_data_vals = _scale_numeric_column(
        data_table,
        data_col_name,
        data_col_vals,
        domain,
    )

    # Validate and get category column
    _, category_col_vals = _validate_and_get_single_column(
        gt,
        category_col,
    )

    color_vals = _get_discrete_colors_from_palette(
        palette=palette, data=category_col_vals, data_table=data_table
    )

    # Apply gt.fmt() to each row individually, so we can access the data_value for that row
    res = gt
    for i in range(len(data_table)):
        data_val = scaled_data_vals[i]
        color_val = color_vals[i]

        res = res.fmt(
            lambda x, data=data_val, fill=color_val: _make_dot_and_bar_svg(
                dot_category_label=x,
                fill=fill,
                bar_val=data,
                svg_height=height,
                svg_width=width,
                font_size=font_size,
            ),
            columns=category_col,
            rows=[i],
        )

    return res


def gt_plt_conf_int(
    gt: GT,
    column: SelectExpr,
    ci_columns: SelectExpr = None,
    ci: float = 0.95,
    width: float = 100,
    height: float = 30,
    dot_color: str = "red",
    dot_border_color: str = "red",
    line_color: str = "royalblue",
    text_color: str = "black",
    font_size: int = 10,
    num_decimals: int = 1,
    # TODO: "none" vs None in text_size
) -> GT:
    """
    Create confidence interval plots in `GT` cells.

    The `gt_plt_conf_int()` function takes an existing `GT` object and adds horizontal confidence
    interval plots to a specified column. Each cell displays a horizontal bar representing the
    confidence interval, with a dot indicating the mean value. Optionally, the lower and upper
    confidence interval bounds can be provided directly, or the function can compute them.

    If `ci_columns` is not provided, the function assumes each cell in `column` contains a list of
    values and computes the confidence interval using a t-distribution.

    Parameters
    ----------
    gt
        A `GT` object to modify.

    column
        The column that contains the mean of the sample. This can either be a single number per row,
        if you have calculated the values ahead of time, or a list of values if you want to
        calculate the confidence intervals.

    ci_columns
        Optional columns representing the left/right confidence intervals of your sample. If `None`,
        the confidence interval will be computed from the data in `column` using a t-distribution.

    ci
        The confidence level to use when computing the interval (if `ci_columns` is `None`).

    width
        The width of the confidence interval plot in pixels. Note that if the width is too narrow,
        some label text may overlap.

    height
        The width of the confidence interval plot in pixels.

    dot_color
        The color of the mean dot.

    dot_border_color
        The color of the border around the mean dot.

    line_color
        The color of the confidence interval bar.

    text_color
        The color of the confidence interval labels.

    font_size
        The size of the text for the confidence interval labels.
        A value of 0 will result in hiding the text.

    num_decimals
        The number of decimals to display when rounding the value of the
        confidence interval labels.

    Returns
    -------
    GT
        A `GT` object with confidence interval plots added to the specified column.

    Examples
    --------
    ```{python}
    import pandas as pd
    from great_tables import GT
    import gt_extras as gte

    df = pd.DataFrame({
        'group': ['A', 'B', 'C'],
        'mean': [5.2, 7.8, 3.4],
        'ci_lower': [3.1, 6.1, 1.8],
        'ci_upper': [7.3, 9.7, 5.0],
        'ci': [5.2, 7.8, 3.4],
    })

    gt = GT(df)
    gt.pipe(
        gte.gt_plt_conf_int,
        column='ci',
        ci_columns=['ci_lower', 'ci_upper'],
        width=120,
    )
    ```

    Alternatively we can pass in a column of lists, and the function will compute the CI's for us.

    ```{python}
    import numpy as np
    np.random.seed(37)

    n_per_group = 50
    groups = ["A", "B", "C"]
    means = [20, 22, 25]
    sds = [10, 16, 10]

    # Create the data
    data = []
    for i, (grp, mean, sd) in enumerate(zip(groups, means, sds)):
        values = np.random.normal(mean, sd, n_per_group)
        data.extend([{"grp": grp, "values": val} for val in values])

    df_raw = pd.DataFrame(data)
    df_summary = (
        df_raw
        .groupby("grp")
        .agg({"values": ["count", "mean", "std", list]})
        .round(3)
    )
    df_summary.columns = ["n", "avg", "sd", "ci"]

    gt = GT(df_summary)
    gt.pipe(
        gte.gt_plt_conf_int,
        column="ci",
        width = 160,
    )
    ```

    Note
    ----
    All confidence intervals are scaled to a common range for visual alignment.
    """

    def _make_conf_int_svg(
        mean: float,
        c1: float,
        c2: float,
        font_size: float,
        min_val: float,
        max_val: float,
        width: float,
        height: float,
        dot_border_color: str,
        line_color: str,
        dot_color: str,
        text_color: str,
        num_decimals: int,
    ) -> str:
        if (
            is_na(gt._tbl_data, mean)
            or is_na(gt._tbl_data, c1)
            or is_na(gt._tbl_data, c2)
        ):
            return f'<div style="display: flex;"><div style="width:{width}px; height:{height}px;"></div></div>'

        span = max_val - min_val

        # Normalize positions to [0, 1] based on global min/max, then scale to width
        c1_pos = ((c1 - min_val) / span) * width
        c2_pos = ((c2 - min_val) / span) * width
        mean_pos = ((mean - min_val) / span) * width

        bar_height = height / 10
        bar_y = height / 2 - bar_height / 2 + font_size / 2

        # Text positioning - place labels above the bar
        label_y = bar_y - (font_size / 2) * 1.2  # 1.2 for padding

        # Dot positioning
        dot_size = height / 5
        dot_y = bar_y - dot_size / 4
        dot_border = height / 20

        # Format the label text
        c1_text = _format_numeric_text(c1, num_decimals)
        c2_text = _format_numeric_text(c2, num_decimals)

        elements = [
            # Confidence interval bar
            Rect(
                x=c1_pos,
                y=bar_y,
                width=c2_pos - c1_pos,
                height=bar_height,
                fill=line_color,
                rx=2,
            ),
            # Mean dot
            Circle(
                cx=mean_pos,
                cy=dot_y + dot_size / 2,
                r=dot_size / 2,
                fill=dot_color,
                stroke=dot_border_color,
                stroke_width=dot_border,
            ),
            # Lower bound label
            Text(
                text=c1_text,
                x=c1_pos,
                y=label_y,
                fill=text_color,
                font_size=font_size,
                text_anchor="start",
                dominant_baseline="central",
            ),
            # Upper bound label
            Text(
                text=c2_text,
                x=c2_pos,
                y=label_y,
                fill=text_color,
                font_size=font_size,
                text_anchor="end",
                dominant_baseline="central",
            ),
        ]

        svg = SVG(width=width, height=height, elements=elements)
        return f'<div style="display: flex;">{svg.as_str()}</div>'

    data_col_name, data_vals = _validate_and_get_single_column(gt, column)

    # must compute the ci ourselves
    if ci_columns is None:
        # Check that all entries are lists or None
        if any(val is not None and not isinstance(val, list) for val in data_vals):
            raise ValueError(
                f"Expected entries in {data_col_name} to be lists or None,"
                "since ci_columns were not given."
            )

        def _compute_mean_and_conf_int(val):
            if val is None or not isinstance(val, list) or len(val) == 0:
                return (None, None, None)
            mean = tmean(val)
            conf_int = t.interval(
                ci,
                len(val) - 1,
                loc=mean,
                scale=sem(val),
            )
            return (mean, conf_int[0], conf_int[1])

        stats = list(map(_compute_mean_and_conf_int, data_vals))
        means, c1_vals, c2_vals = zip(*stats) if stats else ([], [], [])

    # we were given the ci already computed
    else:
        ci_columns_resolved = resolve_cols_c(data=gt, expr=ci_columns)
        if len(ci_columns_resolved) != 2:
            raise ValueError(
                f"Expected 2 ci_columns, instead received {len(ci_columns_resolved)}."
            )

        _, c1_vals = _validate_and_get_single_column(
            gt,
            ci_columns_resolved[0],
        )
        _, c2_vals = _validate_and_get_single_column(
            gt,
            ci_columns_resolved[1],
        )
        means = data_vals

        if any(val is not None and not isinstance(val, (int, float)) for val in means):
            raise ValueError(
                f"Expected all entries in {data_col_name} to be numeric or None,"
                "since ci_columns were given."
            )

    # Compute a global range to ensure conf int bars align
    all_values = [val for val in [*means, *c1_vals, *c2_vals] if val is not None]
    data_min = min(all_values)
    data_max = max(all_values)
    data_range = data_max - data_min

    # Add 10% padding on each side
    padding = data_range * 0.1
    global_min = data_min - padding
    global_max = data_max + padding

    res = gt
    for i in range(len(gt._tbl_data)):
        c1 = c1_vals[i]
        c2 = c2_vals[i]
        mean = means[i]

        res = res.fmt(
            lambda _, c1=c1, c2=c2, mean=mean: _make_conf_int_svg(
                mean=mean,
                c1=c1,
                c2=c2,
                font_size=font_size,
                min_val=global_min,
                max_val=global_max,
                width=width,
                height=height,
                dot_border_color=dot_border_color,
                line_color=line_color,
                dot_color=dot_color,
                text_color=text_color,
                num_decimals=num_decimals,
            ),
            columns=data_col_name,
            rows=[i],
        )

    return res


def gt_plt_dumbbell(
    gt: GT,
    col1: SelectExpr,  # exactly 1 col
    col2: SelectExpr,  # exactly 1 col
    label: str | None = None,
    width: float = 100,
    height: float = 30,
    col1_color: str = "purple",
    col2_color: str = "green",
    bar_color: str = "grey",
    dot_border_color="white",
    font_size: int = 10,
    num_decimals: int = 1,
) -> GT:
    """
    Create dumbbell plots in `GT` cells.

    The `gt_plt_dumbbell()` function takes an existing `GT` object and adds dumbbell plots to
    visualize the difference between two numeric values. Each dumbbell consists of two dots
    (representing values from `col1` and `col2`) connected by a horizontal bar, allowing for
    easy visual comparison between paired values.

    Parameters
    ----------
    gt
        A `GT` object to modify.

    col1
        The column containing the first set of values to plot.

    col2
        The column containing the second set of values to plot.

    label
        Optional label to replace the column name of `col1` in the output table. If `None`, the
        original column name is retained.

    width
        The width of the dumbbell plot in pixels. Note that if the width is too narrow,
        some plot label text may overlap.

    height
        The height of the dumbbell plot in pixels.

    col1_color
        The color of the dots representing values from `col1`.

    col2_color
        The color of the dots representing values from `col2`.

    bar_color
        The color of the horizontal bar connecting the two dots.

    dot_border_color
        The color of the borders around the two dots.

    font_size
        The font size for the value labels displayed above each dot.

    num_decimals
        The number of decimal places to display in the value labels.

    Returns
    -------
    GT
        A `GT` object with dumbbell plots added to the specified columns. The `col2` column is
        hidden from the final table display.

    Examples
    -------
    ```{python}
    import pandas as pd
    from great_tables import GT, html, style, loc
    from great_tables.data import sp500
    import gt_extras as gte

    # Trim the data to December 2008
    df = sp500[["date", "open", "close"]].copy()
    df["date"] = pd.to_datetime(df["date"], errors='coerce')

    dec_2008 = df[
        (df["date"].dt.month == 12) &
        (df["date"].dt.year == 2008)
    ]
    dec_2008 = dec_2008.iloc[::-1].iloc[2:11]

    # Make the Great Table
    gt = (
        GT(dec_2008)
        .tab_source_note(html("Purple: Open<br>Green: Close"))
        .tab_style(
            style=style.text(align="right"),
            locations=[loc.source_notes()]
        )
    )

    gt.pipe(
        gte.gt_plt_dumbbell,
        col1='open',
        col2='close',
        label = "Open to Close ($)",
        num_decimals=0,
        width = 250,
    )

    ```

    Note
    -------
    All dumbbells are scaled to a common range for visual alignment across rows.
    The `col2` column is automatically hidden from the final table display.
    """

    def _make_dumbbell_svg(
        value_1: float,
        value_2: float,
        width: float,
        height: float,
        value_1_color: str,
        value_2_color: str,
        bar_color: str,
        dot_border_color: str,
        max_val: float,
        min_val: float,
        font_size: int,
        num_decimals: int,
    ) -> str:
        if is_na(gt._tbl_data, value_1) or is_na(gt._tbl_data, value_2):
            return f'<div style="display: flex;"><div style="width:{width}px; height:{height}px;"></div></div>'

        # Normalize positions based on global min/max, then scale to width
        span = max_val - min_val
        span = span if span != 0 else 1  # span == 0 is forbidden, causes divide by 0
        pos_1 = ((value_1 - min_val) / span) * width
        pos_2 = ((value_2 - min_val) / span) * width

        # Compute the location of the bar
        bar_left = min(pos_1, pos_2)
        bar_width = abs(pos_2 - pos_1)
        bar_height = height / 10
        bar_y = height / 2 - bar_height / 2 + font_size / 2

        # Compute the locations of the two dots
        dot_radius = bar_height * 1.25
        dot_border = bar_height / 2
        dot_y = bar_y + bar_height / 2

        # Text positioning - labels above dots
        label_y = dot_y - dot_radius - dot_border * 1.2  # 1.2 for padding

        # Format the label text
        value_1_text = _format_numeric_text(value_1, num_decimals)
        value_2_text = _format_numeric_text(value_2, num_decimals)

        elements = [
            # Connecting bar
            Rect(
                x=bar_left,
                y=bar_y,
                width=bar_width,
                height=bar_height,
                fill=bar_color,
                rx=2,
            ),
            # Value 1 dot
            Circle(
                cx=pos_1,
                cy=dot_y,
                r=dot_radius,
                fill=value_1_color,
                stroke=dot_border_color,
                stroke_width=dot_border,
            ),
            # Value 2 dot
            Circle(
                cx=pos_2,
                cy=dot_y,
                r=dot_radius,
                fill=value_2_color,
                stroke=dot_border_color,
                stroke_width=dot_border,
            ),
            # Value 1 label
            Text(
                text=value_1_text,
                x=pos_1,
                y=label_y,
                fill=value_1_color,
                font_size=font_size,
                font_weight="bold",
                text_anchor="middle",
                dominant_baseline="lower",
            ),
            # Value 2 label
            Text(
                text=value_2_text,
                x=pos_2,
                y=label_y,
                fill=value_2_color,
                font_size=font_size,
                font_weight="bold",
                text_anchor="middle",
                dominant_baseline="lower",
            ),
        ]

        svg = SVG(width=width, height=height, elements=elements)
        return f'<div style="display: flex;">{svg.as_str()}</div>'

    col1_name, col1_vals = _validate_and_get_single_column(
        gt,
        col1,
    )
    col2_name, col2_vals = _validate_and_get_single_column(
        gt,
        col2,
    )

    # Check for bad input
    all_values = [val for val in [*col1_vals, *col2_vals] if val is not None]
    if any(val is not None and not isinstance(val, (int, float)) for val in all_values):
        raise ValueError("Expected all entries to be numeric or None.")

    # Compute the global bounds for the column.
    data_min = min(all_values)
    data_max = max(all_values)
    data_range = data_max - data_min

    padding = data_range * 0.1  # Add 10% padding on each side
    global_min = data_min - padding
    global_max = data_max + padding

    res = gt

    for i in range(len(gt._tbl_data)):
        col1_value = col1_vals[i]
        col2_value = col2_vals[i]

        res = res.fmt(
            lambda _, value_1=col1_value, value_2=col2_value: _make_dumbbell_svg(
                value_1=value_1,
                value_2=value_2,
                width=width,
                height=height,
                value_1_color=col1_color,
                value_2_color=col2_color,
                bar_color=bar_color,
                dot_border_color=dot_border_color,
                max_val=global_max,
                min_val=global_min,
                font_size=font_size,
                num_decimals=num_decimals,
            ),
            columns=col1_name,
            rows=[i],
        )

    res = res.cols_hide(col2_name)
    if label is not None:
        res = res.cols_label({col1_name: label})

    return res


def gt_plt_donut(
    gt: GT,
    columns: SelectExpr = None,
    fill: str = "purple",
    size: float = 30,
    stroke_color: str | None = None,
    stroke_width: float = 1,
    show_labels: bool = False,
    label_color: str = "black",
    domain: list[int] | list[float] | None = None,
    keep_columns: bool = False,
) -> GT:
    """
    Create donut charts in `GT` cells.

    The `gt_plt_donut()` function takes an existing `GT` object and adds donut charts to
    specified columns. Each cell value is represented as a portion of a full donut chart,
    with the chart size proportional to the cell's numeric value relative to the column's
    maximum value. The maximum value in the column will display as a full circle.

    Parameters
    ----------
    gt
        A `GT` object to modify.

    columns
        The columns to target. Can be a single column or a list of columns (by name or index).
        If `None`, the donut chart is applied to all numeric columns.

    fill
        The fill color for the donut chart segments.

    size
        The diameter of the donut chart in pixels.

    stroke_color
        The color of the border around the donut chart. If `None`, no stroke will be drawn,
        except for the case of the 0 value.

    stroke_width
        The width of the border stroke in pixels.

    show_labels
        Whether or not to show labels on the donut charts.

    label_color
        The color of text labels on the donut charts (when `show_labels` is `True`).

    domain
        The domain of values to use for scaling. This can be a list of floats or integers.
        If `None`, the domain is automatically set to `[0, max(column_values)]`.

    keep_columns
        Whether to keep the original column values. If this flag is `True`, the plotted values will
        be duplicated into a new column with the string " plot" appended to the end of the column
        name. See [`gt_duplicate_column()`](https://posit-dev.github.io/gt-extras/reference/gt_duplicate_column)
        for more details.

    Returns
    -------
    GT
        A `GT` object with donut charts added to the specified columns.

    Examples
    --------

    ```{python}
    from great_tables import GT
    from great_tables.data import gtcars
    import gt_extras as gte

    gtcars_mini = gtcars.loc[
        9:17,
        ["model", "mfr", "year", "hp", "hp_rpm", "trq", "trq_rpm", "mpg_c", "mpg_h"]
    ]

    gt = (
        GT(gtcars_mini, rowname_col="model")
        .tab_stubhead(label="Car")
        .cols_align("center")
        .cols_align("left", columns="mfr")
    )

    gt.pipe(
        gte.gt_plt_donut,
        columns=["hp", "hp_rpm", "trq", "trq_rpm", "mpg_c", "mpg_h"],
        size=40,
        fill="steelblue"
    )
    ```

    Note
    --------
    Each column's donut charts are scaled independently based on that column's min/max values.
    A value equal to the column maximum will display as a full circle (360 degrees).
    """

    # Allow the user to hide the stroke
    if stroke_color is None:
        stroke_color = "transparent"
        stroke_width = 0

    def _make_pie_svg(
        scaled_val: float,
        original_val: int | float,
        fill: str,
        size: float,
        stroke_color: str,
        stroke_width: float,
        show_labels: bool,
        label_color: str,
    ) -> str:
        if is_na(gt._tbl_data, original_val):
            return f'<div style="display: flex;"><div style="width:{size}px; height:{size}px;"></div></div>'

        elements = []
        svg_style = ""

        radius = size / 2
        center_x = center_y = radius
        inner_radius = radius * 0.4

        # Calculate the angle in radians (0 to 2π)
        angle = min(scaled_val * 2 * math.pi, 2 * math.pi)

        # Outer arc start/end points
        outer_start_x = center_x
        outer_start_y = stroke_width / 2
        outer_end_x = center_x + (radius - stroke_width / 2) * math.sin(angle)
        outer_end_y = center_y - (radius - stroke_width / 2) * math.cos(angle)

        # Inner arc start/end points
        inner_start_x = center_x
        inner_start_y = center_y - inner_radius
        inner_end_x = center_x + inner_radius * math.sin(angle)
        inner_end_y = center_y - inner_radius * math.cos(angle)

        # Determine if we need a large arc (> 180 degrees)
        large_arc = angle > math.pi

        if scaled_val <= 0:
            stroke_width = max(stroke_width, 1)
            if stroke_color == "transparent":
                stroke_color = "black"

            # Draw empty donut with just stroke/outline
            outer_circle = Circle(
                cx=center_x,
                cy=center_y,
                r=radius - stroke_width / 2,
                fill="transparent",
                stroke=stroke_color,
                stroke_width=stroke_width,
                stroke_dasharray=3,
            )
            elements.append(outer_circle)

        else:
            # For full circle (angle >= 2π), we need special handling to avoid degenerate arcs
            if angle >= 2 * math.pi - 0.001:
                # Create full donut using two semicircular arcs with evenodd fill rule
                path_commands = [
                    # Outer circle - first semicircle (top to bottom)
                    MoveTo(center_x, stroke_width / 2),  # Top
                    Arc(
                        radius - stroke_width / 2,
                        radius - stroke_width / 2,
                        0,
                        False,  # small arc
                        True,  # clockwise
                        center_x,
                        center_y + (radius - stroke_width / 2),  # Bottom
                    ),
                    # Outer circle - second semicircle (bottom to top)
                    Arc(
                        radius - stroke_width / 2,
                        radius - stroke_width / 2,
                        0,
                        False,  # small arc
                        True,  # clockwise
                        center_x,
                        stroke_width / 2,  # Back to top
                    ),
                    ClosePath(),
                    # Inner circle - first semicircle (top to bottom, counter-clockwise)
                    MoveTo(center_x, center_y - inner_radius),  # Top of inner circle
                    Arc(
                        inner_radius,
                        inner_radius,
                        0,
                        False,  # small arc
                        False,  # counter-clockwise
                        center_x,
                        center_y + inner_radius,  # Bottom of inner circle
                    ),
                    # Inner circle - second semicircle (bottom to top, counter-clockwise)
                    Arc(
                        inner_radius,
                        inner_radius,
                        0,
                        False,  # small arc
                        False,  # counter-clockwise
                        center_x,
                        center_y - inner_radius,  # Back to top
                    ),
                    ClosePath(),
                ]
                svg_style = "fill-rule:true;"
            else:
                # Partial donut using path: outer arc -> line to inner end -> inner arc (reverse) -> close
                path_commands = [
                    MoveTo(outer_start_x, outer_start_y),  # Start at outer edge
                    Arc(
                        radius - stroke_width / 2,
                        radius - stroke_width / 2,
                        0,
                        large_arc,
                        True,
                        outer_end_x,
                        outer_end_y,
                    ),  # Outer arc
                    LineTo(inner_end_x, inner_end_y),  # Line to inner edge
                    Arc(
                        inner_radius,
                        inner_radius,
                        0,
                        large_arc,
                        False,  # Reverse direction for inner arc
                        inner_start_x,
                        inner_start_y,
                    ),  # Inner arc (reverse)
                    ClosePath(),
                ]

            path = Path(
                d=path_commands,
                fill=fill,
                stroke=stroke_color,
                stroke_width=stroke_width,
            )
            elements.append(path)

        # Add label if requested
        if show_labels and not is_na(gt._tbl_data, original_val):
            label_text = str(original_val)

            # Position label in the center of the donut
            label_x = center_x
            label_y = center_y
            font_size = size * 0.15

            text = Text(
                text=label_text,
                x=label_x,
                y=label_y,
                fill=label_color,
                font_size=font_size,
                text_anchor="middle",
                dominant_baseline="central",
                font_weight="bold",
            )
            elements.append(text)

        svg = SVG(width=size, height=size, elements=elements, style=svg_style)
        return f'<div style="display: flex;">{svg.as_str()}</div>'

    # Get names of columns
    columns_resolved = resolve_cols_c(data=gt, expr=columns)

    res = gt
    for column in columns_resolved:
        # Validate this is a single column and get values
        col_name, col_vals = _validate_and_get_single_column(
            gt,
            column,
        )

        scaled_vals = _scale_numeric_column(
            res._tbl_data,
            col_name,
            col_vals,
            domain,
        )

        # The location of the plot column will be right after the original column
        if keep_columns:
            res = gt_duplicate_column(
                res,
                col_name,
                after=col_name,
                append_text=" plot",
            )
            col_name = col_name + " plot"

        # Apply the scaled value for each row, so the donut is proportional
        for i, scaled_val in enumerate(scaled_vals):
            res = res.fmt(
                lambda original_val, scaled_val=scaled_val: _make_pie_svg(
                    original_val=original_val,
                    scaled_val=scaled_val,
                    fill=fill,
                    size=size,
                    stroke_color=stroke_color,
                    stroke_width=stroke_width,
                    show_labels=show_labels,
                    label_color=label_color,
                ),
                columns=col_name,
                rows=[i],
            )

    return res


def gt_plt_winloss(
    gt: GT,
    column: SelectExpr,
    width: float = 80,
    height: float = 30,
    win_color: str = "blue",
    loss_color: str = "red",
    tie_color: str = "grey",
    shape: Literal["pill", "square"] = "pill",
    spacing: float = 2,
) -> GT:
    """
    Create win/loss charts in `GT` cells.

    The `gt_plt_winloss()` function takes an existing `GT` object and adds win/loss sparkline
    charts to a specified column. Each cell displays a series of small vertical bars representing
    individual game outcomes, This visualization is useful for showing performance streaks and
    patterns over time. All win/loss charts are scaled to accommodate the longest sequence in the
    column, ensuring consistent bar spacing across all rows.

    Wins must be represented as `1`, ties as `0.5`, and losses as `0`.
    Invalid values (not `0`, `0.5`, or `1`) are skipped.

    Parameters
    ----------
    gt
        A `GT` object to modify.

    column
        The column containing lists of win/loss/tie values. Each cell should contain a list where:
        `1` represents a win, `0` represents a loss, and `0.5` represents a tie.
        Values that are not listed above are skipped.

    width
        The width of the win/loss chart in pixels.

    height
        The height of the win/loss chart in pixels.

    win_color
        The color for bars representing wins.

    loss_color
        The color for bars representing losses.

    tie_color
        The color for bars representing ties.

    shape
        The shape style of the bars. Options are `"pill"` for taller bars or `"square"` for
        stockier, nearly square bars.

    spacing
        The horizontal gap, in pixels, between each bar. Note that if the spacing is too large, it
        may obstruct the bars from view.

    Returns
    -------
    GT
        A `GT` object with win/loss charts added to the specified column.

    Examples
    --------
    First, let's make a table with randomly generated data:

    ```{python}
    from great_tables import GT, md
    import gt_extras as gte
    import pandas as pd

    df = pd.DataFrame(
        {
            "Team": ["Liverpool", "Chelsea", "Man City"],
            "10 Games": [
                [1, 1, 0, 1, 0.5, 1, 0, 1, 1, 0],
                [0, 0, 1, 0, 1, 1, 1, 0, 1, 1],
                [0.5, 1, 0.5, 0, 1, 0, 1, 0.5, 1, 0],
            ],
        }
    )

    gt = GT(df)

    gt.pipe(
        gte.gt_plt_winloss,
        column="10 Games",
        win_color="green",
    )
    ```

    Let's do a more involved example using NFL season data from 2016.

    ```{python}
    #| code-fold: true
    #| code-summary: Show the setup Code

    # Load the NFL data
    df = pd.read_csv("../assets/games.csv")
    season_2016 = df[(df["season"] == 2016) & (df["game_type"] == "REG")].copy()

    def get_team_results(games_df):
        results = {}

        for _, game in games_df.iterrows():
            away_team = game["away_team"]
            home_team = game["home_team"]
            away_score = game["away_score"]
            home_score = game["home_score"]

            if away_team not in results:
                results[away_team] = []
            if home_team not in results:
                results[home_team] = []

            if away_score > home_score:
                results[away_team].append(1)
                results[home_team].append(0)
            elif home_score > away_score:
                results[home_team].append(1)
                results[away_team].append(0)
            else:
                results[away_team].append(0.5)
                results[home_team].append(0.5)

        return results

    team_results = get_team_results(season_2016)
    winloss_df = pd.DataFrame(
        [{"Team": team, "Games": results} for team, results in team_results.items()]
    )

    winloss_df = (
        winloss_df
        .sort_values("Team")
        .reset_index(drop=True)
        .head(10)
    )
    ```

    Now that we've loaded the real-world data, let's see how we can use `gt_plt_winloss()`.

    ```{python}
    gt = (
        GT(winloss_df)
        .tab_header(
            title="2016 NFL Season",
        )
        .tab_source_note(
            md(
                '<span style="float: right;">Source: [Lee Sharpe, nflverse](https://github.com/nflverse/nfldata)</span>'
            )
        )
        .cols_align("left", columns="Games")
    )

    gt.pipe(
        gte.gt_plt_winloss,
        column="Games",
    )
    ```
    """

    def _make_winloss_svg(
        values: list[float],
        max_length: int,
        width: float,
        height: float,
        win_color: str,
        loss_color: str,
        tie_color: str,
        shape: Literal["pill", "square"],
        spacing: float,
    ) -> str:
        if len(values) == 0:
            return f'<div style="display: flex;"><div style="width:{width}px; height:{height}px;"></div></div>'

        available_width = width - (max_length) * spacing
        bar_width = available_width / max_length
        win_bar_height = height * 0.2 if shape == "square" else height * 0.4

        elements = []

        for i, value in enumerate(values):
            if is_na(gt._tbl_data, value):
                continue

            if value == 1:  # Win
                color = win_color
                bar_y = height * 0.2
                bar_height = win_bar_height
            elif value == 0.5:  # Tie
                color = tie_color
                bar_y = height * 0.4
                bar_height = height * 0.2
            elif value == 0:  # Loss
                color = loss_color
                bar_y = height * 0.8 - win_bar_height
                bar_height = win_bar_height
            else:
                warnings.warn(
                    f"Invalid value '{value}' encountered in win/loss data. Skipping.",
                    category=UserWarning,
                )
                continue

            bar_x = i * (bar_width + spacing)
            border_radius = 0.5 if shape == "square" else 2

            bar_rect = Rect(
                x=bar_x,
                y=bar_y,
                width=bar_width,
                height=bar_height,
                fill=color,
                rx=border_radius,
            )
            elements.append(bar_rect)

        svg = SVG(width=width, height=height, elements=elements)
        return f'<div style="display: flex;">{svg.as_str()}</div>'

    res = gt
    _, col_vals = _validate_and_get_single_column(gt, expr=column)
    max_length = max(len(entry) for entry in col_vals)

    if spacing * max_length >= width:
        warnings.warn(
            "Spacing is too large relative to the width. No bars will be displayed.",
            category=UserWarning,
        )

    # I don't have to loop like with the others since I dont need to access other columns
    res = res.fmt(
        lambda x: _make_winloss_svg(
            x,
            max_length=max_length,
            width=width,
            height=height,
            win_color=win_color,
            loss_color=loss_color,
            tie_color=tie_color,
            shape=shape,
            spacing=spacing,
        ),
        columns=column,
    )

    return res


def gt_plt_bar_stack(
    gt: GT,
    column: SelectExpr,
    labels: list[str] | None = None,
    width: float = 100,
    height: float = 30,
    palette: list[str] | str | None = None,
    font_size: int = 10,
    spacing: float = 2,
    num_decimals: int = 0,
    scale_type: Literal["relative", "absolute"] = "relative",
) -> GT:
    """
    Create stacked horizontal bar plots in `GT` cells.

    The `gt_plt_bar_stack()` function takes an existing `GT` object and adds stacked horizontal bar
    charts to a specified column. Each cell displays a series of horizontal bars whose lengths are
    proportional to the values in the list. The scaling of the bars can be controlled using the
    `scale_type` - see below for more info.

    Parameters
    ----------
    gt
        A `GT` object to modify.

    column
        The column containing lists of numeric values to represent as stacked horizontal bars. Each
        cell should contain a list of numeric values.

    labels
        Optional labels for the bars. If provided, these labels will be displayed in the column
        header, with each label corresponding to a color in the palette.

    width
        The total width of the stacked bar plot in pixels. If `scale_type = "absolute"`, this
        value will determine the width of the maximum length bar plot.

    height
        The height of the stacked bar plot in pixels.

    palette
        The color palette to use for the bars. This can be a list of colors
        (e.g., `["#FF0000"`, `"#00FF00"`, `"#0000FF"` `]`) or a named palette (e.g., `"viridis"`).
        If `None`, a default palette will be used.

    font_size
        The font size for the text labels displayed on the bars.

    spacing
        The horizontal gap, in pixels, between each bar. If the spacing is too large relative to
        the width, a warning will be issued, and no bars will be displayed.

    num_decimals
        The number of decimal places to display in the text labels on the bars.

    scale_type
        Determines how the bars are scaled. Options are `"relative"` (bars are scaled relative to
        the sum of the values in each cell) and `"absolute"` (bars are scaled relative to the
        maximum value across all rows).

    Returns
    -------
    GT
        A `GT` object with stacked horizontal bar plots added to the specified column.

    Examples
    --------
    ```{python}
    import pandas as pd
    from great_tables import GT
    import gt_extras as gte

    df = pd.DataFrame({
        "x": ["Example A", "Example B", "Example C"],
        "col": [
            [10, 40, 50],
            [30, 30, 40],
            [50, 20, 30],
        ],
    })

    gt = GT(df)

    gt.pipe(
        gte.gt_plt_bar_stack,
        column="col",
        palette=["red", "grey", "black"],
        labels=["Group 1", "Group 2", "Group 3"],
        width=200,
    )
    ```

    If the absolute sum of each row varies, we can treat the rows as portions of a whole.

    ```{python}
    df = pd.DataFrame({
        "x": ["Example A", "Example B", "Example C"],
        "col": [
            [10, 20, 50],
            [30, 30],
            [50, 10, 10],
        ],
    })

    gt = GT(df)

    gt.pipe(
        gte.gt_plt_bar_stack,
        column="col",
        labels=["Group 1", "Group 2", "Group 3"],
        width=200,
        scale_type="relative",
    )
    ```

    Or we can treat them as absolute values.

    ```{python}
    df = pd.DataFrame({
        "x": ["Example A", "Example B", "Example C"],
        "col": [
            [10, 20, 50],
            [30, 30],
            [50, 10, 10],
        ],
    })

    gt = GT(df)

    gt.pipe(
        gte.gt_plt_bar_stack,
        column="col",
        labels=["Group 1", "Group 2", "Group 3"],
        width=200,
        scale_type="absolute",
    )
    ```

    Note
    -------
    Values of `0` will not be displayed in the plots.
    """

    def _make_bar_stack_svg(
        values: list[float],
        max_sum: float,
        width: float,
        height: float,
        colors: list[str],
        spacing: float,
        num_decimals: int,
        font_size: int,
        scale_type: Literal["relative", "absolute"],
    ) -> str:
        if not values:
            return f'<div style="display: flex;"><div style="width:{width}px; height:{height}px;"></div></div>'

        non_na_vals = [val if not is_na(gt._tbl_data, val) else 0 for val in values]
        # Count how many values will be displayed in the chart
        len_non_zero_values = sum(1 for val in non_na_vals if val != 0)

        if scale_type == "absolute":
            total = max_sum
        else:
            total = sum(non_na_vals)

        # Avoid div by 0
        if total == 0:
            total = 1

        normalized_values = [val / total for val in non_na_vals]
        available_width = width - (len_non_zero_values - 1) * spacing
        if available_width <= 0:
            warnings.warn(
                "Spacing is too large relative to the width. No bars will be displayed.",
                category=UserWarning,
            )
            return f'<div style="display: flex;"><div style="width:{width}px; height:{height}px;"></div></div>'

        elements = []
        current_left = 0

        for i, value in enumerate(normalized_values):
            if value == 0:
                continue

            bar_width = available_width * value
            color = colors[i % len(colors)]

            # Create the bar rectangle
            bar_rect = Rect(
                x=current_left,
                y=0,
                width=bar_width,
                height=height,
                fill=color,
            )
            elements.append(bar_rect)

            # Create the label text
            label = f"{non_na_vals[i]:.{num_decimals}f}"
            label_text = Text(
                text=label,
                x=current_left + bar_width / 2,  # Center horizontally in the bar
                y=height / 2,  # Center vertically
                fill=_ideal_fgnd_color(_html_color([color])[0]),
                font_size=font_size,
                text_anchor="middle",
                dominant_baseline="central",
            )
            elements.append(label_text)

            current_left += bar_width + spacing

        svg = SVG(width=width, height=height, elements=elements)
        return f'<div style="display: flex;">{svg.as_str()}</div>'

    # Throw if `scale_type` is not one of the allowed values
    if scale_type not in ["relative", "absolute"]:
        raise ValueError("Scale_type must be either 'relative' or 'absolute'")

    col_name, col_vals = _validate_and_get_single_column(gt, expr=column)
    cleaned_col_vals = [
        [val for val in col if val is not None and not is_na(gt._tbl_data, val)]
        for col in col_vals
        if col is not None
    ]

    max_sum = max(sum(col) for col in cleaned_col_vals if col is not None)
    max_num_values = max(len(col) for col in cleaned_col_vals)

    # If user passes a list, accept those colors, otherwise use palette functionality.
    if isinstance(palette, list) and len(palette) >= max_num_values:
        color_list = palette
    else:
        color_list = _get_discrete_colors_from_palette(
            palette=palette,
            data=[i for i in range(max_num_values)],
            data_table=gt._tbl_data,
        )

    res = gt
    res = res.fmt(
        lambda x: _make_bar_stack_svg(
            x,
            max_sum=max_sum,
            width=width,
            height=height,
            colors=color_list,
            spacing=spacing,
            font_size=font_size,
            num_decimals=num_decimals,
            scale_type=scale_type,
        ),
        columns=column,
    )

    if labels is not None:
        label_html = [
            f'<span style="color:{color}">{name}</span>'
            for name, color in zip(labels, color_list)
        ]
        label_html = " | ".join(label_html)
        res = res.cols_align(columns=col_name, align="center").cols_label(
            {col_name: html(label_html)}
        )

    return res


def gt_plt_bar_pct(
    gt: GT,
    column: SelectExpr,
    height: int = 16,
    width: int = 100,
    fill: str = "purple",
    background: str = "#e1e1e1",
    autoscale: bool = True,
    labels: bool = False,
    label_cutoff: float = 0.4,
    decimals: int = 1,
    font_style: Literal["oblique", "italic", "normal"] = "normal",
    font_size: int = 10,
):
    """
    Create horizontal bar plots in percentage in `GT` cells.

    The `gt_plt_bar_pct()` function takes an existing `GT` object and adds
    horizontal bar plots via native HTML. By default, values are normalized
    as a percentage of the maximum value in the specified column. If the values
    already represent percentages (i.e., between 0–100), you can disable this
    behavior by setting `autoscale=False`.

    Parameters
    ----------
    gt
        A `GT` object to modify.

    column
        The column to target.

    height
        The height of the bar plot in pixels.

    width
        The width of the maximum bar in pixels.

    fill
        The fill color for the bars.

    background
        The background filling color for the bars. Defaults to `#e1e1e1`.

    autoscale
        Indicates whether the function should automatically scale the values.
        If `True`, values will be divided by the column's maximum and multiplied by 100.
        If `False`, the values are assumed to already be scaled appropriately.

    labels
        `True`/`False` logical representing if labels should be plotted. Defaults
        to `False`, meaning that no value labels will be plotted.

    label_cutoff
        A number, 0 to 1, representing where to set the inside/outside label
        boundary. Defaults to 0.40 (40%) of the column's maximum value. If the
        value in that row is less than the cutoff, the label will be placed
        outside the bar; otherwise, it will be placed within the bar. This
        interacts with the overall width of the bar, so if you are not happy with
        the placement of the labels, you may try adjusting the `width` argument as
        well.

    decimals
        A number representing how many decimal places to be used in label
        rounding.

    font_style
        The font style for the text labels displayed on the bars. Options are
        `"oblique"`, `"italic"`, or `"normal"`.

    font_size
        The font size for the text labels displayed on the bars.

    Returns
    -------
    GT
        A `GT` object with horizontal bar plots added to the specified columns.

    Examples
    --------
    The `autoscale` parameter is perhaps the most important in the `gt_plt_bar_pct()` function.
    This example demonstrates the difference between `autoscale=True` and `autoscale=False` using column `x`:

    * **When `autoscale=True`**:
    The function scales the values relative to the maximum in the column.
    For example, `[10, 20, 30, 40]` becomes `[25%, 50%, 75%, 100%]`,
    which are used for both bar lengths and labels.

    * **When `autoscale=False`**:
    The values are assumed to already represent percentages.
    The function uses them as-is — e.g., `[10%, 20%, 30%, 40%]`,
    which are directly reflected in both the bar lengths and labels.

    ```{python}
    import polars as pl
    from great_tables import GT
    import gt_extras as gte

    df = pl.DataFrame({"x": [10, 20, 30, 40]}).with_columns(
        pl.col("x").alias("autoscale_on"),
        pl.col("x").alias("autoscale_off"),
    )

    gt = GT(df)

    (
        gt.pipe(
            gte.gt_plt_bar_pct,
            column=["autoscale_on"],
            autoscale=True,
            labels=True,
            fill="green",
        ).pipe(
            gte.gt_plt_bar_pct,
            column=["autoscale_off"],
            autoscale=False,
            labels=True,
        )
    )
    ```
    Finally, label colors are automatically adjusted based on the `fill` and `background` parameters
    to ensure optimal readability.
    """

    def _not_na(val, tbl_data) -> bool:
        return val is not None and not is_na(tbl_data, val)

    def _is_na(val, tbl_data) -> bool:
        return not _not_na(val, tbl_data)

    def _is_effective_int(val) -> bool:
        return isinstance(val, int) or (isinstance(val, float) and val.is_integer())

    if not (0 <= label_cutoff <= 1):
        raise ValueError("Label_cutoff must be a number between 0 and 1.")

    if font_style not in ["bold", "italic", "normal"]:
        raise ValueError("Font_style must be one of 'bold', 'italic', or 'normal'.")

    # Helper function to make the individual bars

    def _make_bar_pct_svg(
        # original_val: int | float,
        scaled_val: int | float,
        height: int,
        width: int,
        fill: str,
        background: str,
        # autoscale: bool,
        labels: bool,
        label_cutoff: float,
        decimals: int,
        font_style: Literal["oblique", "italic", "normal"],
        font_size: int,
    ) -> str:
        elements = []
        if _is_na(scaled_val, tbl_data):
            outer_rect = Rect(
                x=0,
                y=0,
                width=Length(width, "%"),
                height=Length(height, "px"),
                fill="transparent",
            )
            elements.append(outer_rect)
        else:
            outer_rect = Rect(
                x=0,
                y=0,
                width=Length(width, "px"),
                height=Length(height, "px"),
                fill=background,
            )
            elements.append(outer_rect)

            _width = width * scaled_val * 0.01

            inner_rect = Rect(
                x=0,
                y=0,
                width=Length(_width, "px"),
                height=Length(height, "px"),
                fill=fill,
            )
            elements.append(inner_rect)

            if labels:
                padding = 5.0

                if _width < (label_cutoff * 100):
                    _x = _width + padding
                    _fill = _ideal_fgnd_color(_html_color([background])[0])
                else:
                    _x = padding
                    _fill = _ideal_fgnd_color(_html_color([fill])[0])

                _decimals = decimals
                if _is_effective_int(scaled_val):
                    _decimals = 0
                _text = f"{scaled_val:.{_decimals}f}%"

                inner_text = Text(
                    text=_text,
                    x=Length(_x, "px"),
                    y=Length(height / 2, "px"),
                    fill=_fill,
                    font_size=Length(font_size, "px"),
                    font_style=font_style,
                    text_anchor="start",
                    dominant_baseline="central",
                )
                elements.append(inner_text)

        canvas = SVG(
            width=Length(width, "px"),
            height=Length(height, "px"),
            elements=elements,
        )
        return f'<div style="display: flex;">{canvas.as_str()}</div>'

    def _make_bar_pct(scaled_val: int) -> str:
        return _make_bar_pct_svg(
            # original_val=original_val,
            scaled_val=scaled_val,
            height=height,
            width=width,
            fill=fill,
            background=background,
            # autoscale=autoscale,
            labels=labels,
            label_cutoff=label_cutoff,
            decimals=decimals,
            font_style=font_style,
            font_size=font_size,
        )

    _, col_vals = _validate_and_get_single_column(gt, expr=column)
    tbl_data = gt._tbl_data
    if all(_is_na(val, tbl_data) for val in col_vals):
        raise ValueError("All values in the column are None.")

    max_x = max(val for val in col_vals if _not_na(val, tbl_data))

    scaled_vals = col_vals
    if autoscale:
        scaled_vals = [
            val if _is_na(val, tbl_data) else (val / max_x * 100) for val in col_vals
        ]

    res = gt
    # Apply the scaled value for each row, so the bar is proportional
    for i, scaled_val in enumerate(scaled_vals):
        res = res.fmt(
            lambda _, scaled_val=scaled_val: _make_bar_pct(scaled_val=scaled_val),
            columns=column,
            rows=[i],
        )
    return res


########### Helper functions that get reused across plots ###########


# Helper function to make the individual bars
def _make_bar_svg(
    scaled_val: float,
    original_val: int | float,
    fill: str,
    bar_height: float,
    height: float,
    width: float,
    stroke_color: str,
    show_labels: bool,
    label_color: str | None,
) -> SVG:
    text = ""
    if show_labels:
        text = str(original_val)

    elements = [
        Rect(
            x=0,
            y=Length((height - bar_height) / 2, "px"),
            width=Length(width * scaled_val, "px"),
            height=Length(bar_height, "px"),
            fill=fill,
            # onmouseover="this.style.fill= 'blue';",
            # onmouseout=f"this.style.fill='{fill}';",
        ),
        Text(
            text=text,
            x=Length((width * scaled_val) * 0.98, "px"),
            y=Length(height / 2, "px"),
            fill=label_color,
            font_size=bar_height * 0.6,
            text_anchor="end",
            dominant_baseline="central",
        ),
        Line(
            x1=0,
            x2=0,
            y1=0,
            y2=Length(height, "px"),
            stroke_width=Length(height / 10, "px"),
            stroke=stroke_color,
        ),
    ]

    return SVG(width=width, height=height, elements=elements)



from __future__ import annotations

import warnings

from great_tables._data_color.base import (
    _get_domain_factor,
    _html_color,
    _rescale_factor,
)
from great_tables._data_color.constants import ALL_PALETTES, DEFAULT_PALETTE
from great_tables._data_color.palettes import GradientPalette
from great_tables._tbl_data import TblData

__all__ = ["_get_discrete_colors_from_palette"]


def _get_discrete_colors_from_palette(
    palette: list[str] | str | None,
    data: list,
    data_table: TblData,
) -> list[str]:
    palette = _get_palette(palette)

    # Rescale the category column for the purpose of assigning colors to each dot
    domain_factor = _get_domain_factor(df=data_table, vals=data)
    scaled_vals = _rescale_factor(
        df=data_table,
        vals=data,
        domain=domain_factor,  # type: ignore
        palette=palette,
    )

    # Create a color scale function from the palette
    color_scale_fn = GradientPalette(colors=palette)

    # Call the color scale function on the scaled categoy values to get a list of colors
    color_vals = color_scale_fn(scaled_vals)

    for i, c in enumerate(color_vals):
        if c is None:
            color_vals[i] = "transparent"
            warnings.warn(
                "A color value is None and has been coerced to 'transparent'",
                UserWarning,
            )

    # This is redundant but satisfies the type-checker
    color_vals = [c for c in color_vals if c is not None]

    return color_vals


def _get_palette(palette: list[str] | str | None) -> list[str]:
    # If palette is not provided, use a default palette
    if palette is None:
        palette = DEFAULT_PALETTE

    # Otherwise get the palette from great_tables._data_color
    elif isinstance(palette, str):
        palette = ALL_PALETTES.get(palette, [palette])

    # Standardize values in `palette` to hexadecimal color values
    palette = _html_color(colors=palette)

    return palette
